# -*- coding: utf-8 -*-
# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
# Copyright 2014 Aaron Whitehouse <aaron@whitehouse.kiwi.nz>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

from __future__ import print_function
from future import standard_library
standard_library.install_aliases()

import types
import io
import unittest
import duplicity.path

from duplicity.selection import *  # @UnusedWildImport
from duplicity.lazy import *  # @UnusedWildImport
from . import UnitTestCase
from mock import patch


class MatchingTest(UnitTestCase):
    u"""Test matching of file names against various selection functions"""
    def setUp(self):
        super(MatchingTest, self).setUp()
        self.unpack_testfiles()
        self.root = Path(u"testfiles/select")
        self.Select = Select(self.root)

    def makeext(self, path):
        return self.root.new_index(tuple(path.encode().split(b"/")))

    def testRegexp(self):
        u"""Test regular expression selection func"""
        sf1 = self.Select.regexp_get_sf(u".*\\.py", 1)
        assert sf1(self.makeext(u"1.py")) == 1
        assert sf1(self.makeext(u"usr/foo.py")) == 1
        assert sf1(self.root.append(u"1.doc")) is None

        sf2 = self.Select.regexp_get_sf(u"hello", 0)
        assert sf2(Path(u"hello")) == 0
        assert sf2(Path(u"foohello_there")) == 0
        assert sf2(Path(u"foo")) is None

    def test_tuple_include(self):
        u"""Test include selection function made from a regular filename"""
        self.assertRaises(FilePrefixError, self.Select.glob_get_normal_sf,
                          u"foo", 1)

        sf2 = self.Select.glob_get_sf(u"testfiles/select/usr/local/bin/", 1)

        with patch(u"duplicity.path.ROPath.isdir") as mock_isdir:
            mock_isdir.return_value = True
            # Can't pass the return_value as an argument to patch, i.e.:
            # with patch("duplicity.path.ROPath.isdir", return_value=True):
            # as build system's mock is too old to support it.

            self.assertEqual(sf2(self.makeext(u"usr")), 2)
            self.assertEqual(sf2(self.makeext(u"usr/local")), 2)
            self.assertEqual(sf2(self.makeext(u"usr/local/bin")), 1)
            self.assertEqual(sf2(self.makeext(u"usr/local/doc")), None)
            self.assertEqual(sf2(self.makeext(u"usr/local/bin/gzip")), 1)
            self.assertEqual(sf2(self.makeext(u"usr/local/bingzip")), None)

    def test_tuple_exclude(self):
        u"""Test exclude selection function made from a regular filename"""
        self.assertRaises(FilePrefixError, self.Select.glob_get_normal_sf,
                          u"foo", 0)

        sf2 = self.Select.glob_get_sf(u"testfiles/select/usr/local/bin/", 0)

        with patch(u"duplicity.path.ROPath.isdir") as mock_isdir:
            mock_isdir.return_value = True

            assert sf2(self.makeext(u"usr")) is None
            assert sf2(self.makeext(u"usr/local")) is None
            assert sf2(self.makeext(u"usr/local/bin")) == 0
            assert sf2(self.makeext(u"usr/local/doc")) is None
            assert sf2(self.makeext(u"usr/local/bin/gzip")) == 0
            assert sf2(self.makeext(u"usr/local/bingzip")) is None

    def test_glob_star_include(self):
        u"""Test a few globbing patterns, including **"""
        sf1 = self.Select.glob_get_sf(u"**", 1)
        assert sf1(self.makeext(u"foo")) == 1
        assert sf1(self.makeext(u"")) == 1

        sf2 = self.Select.glob_get_sf(u"**.py", 1)
        assert sf2(self.makeext(u"foo")) == 2
        assert sf2(self.makeext(u"usr/local/bin")) == 2
        assert sf2(self.makeext(u"what/ever.py")) == 1
        assert sf2(self.makeext(u"what/ever.py/foo")) == 1

    def test_glob_star_exclude(self):
        u"""Test a few glob excludes, including **"""
        sf1 = self.Select.glob_get_sf(u"**", 0)
        assert sf1(self.makeext(u"/usr/local/bin")) == 0

        sf2 = self.Select.glob_get_sf(u"**.py", 0)
        assert sf2(self.makeext(u"foo")) is None
        assert sf2(self.makeext(u"usr/local/bin")) is None
        assert sf2(self.makeext(u"what/ever.py")) == 0
        assert sf2(self.makeext(u"what/ever.py/foo")) == 0

    def test_simple_glob_double_asterisk(self):
        u"""test_simple_glob_double_asterisk - primarily to check that the defaults used by the error tests work"""
        assert self.Select.glob_get_normal_sf(u"**", 1)

    def test_glob_sf_exception(self):
        u"""test_glob_sf_exception - see if globbing errors returned"""
        self.assertRaises(GlobbingError, self.Select.glob_get_normal_sf,
                          u"testfiles/select/hello//there", 1)

    def test_file_prefix_sf_exception(self):
        u"""test_file_prefix_sf_exception - see if FilePrefix error is returned"""
        # These should raise a FilePrefixError because the root directory for the selection is "testfiles/select"
        self.assertRaises(FilePrefixError,
                          self.Select.glob_get_sf, u"testfiles/whatever", 1)
        self.assertRaises(FilePrefixError,
                          self.Select.glob_get_sf, u"testfiles/?hello", 0)

    def test_scan(self):
        u"""Tests what is returned for selection tests regarding directory scanning"""
        select = Select(Path(u"/"))

        assert select.glob_get_sf(u"**.py", 1)(Path(u"/")) == 2
        assert select.glob_get_sf(u"**.py", 1)(Path(u"foo")) == 2
        assert select.glob_get_sf(u"**.py", 1)(Path(u"usr/local/bin")) == 2
        assert select.glob_get_sf(u"/testfiles/select/**.py", 1)(Path(u"/testfiles/select")) == 2
        assert select.glob_get_sf(u"/testfiles/select/test.py", 1)(Path(u"/testfiles/select")) == 2
        assert select.glob_get_normal_sf(u"/testfiles/se?ect/test.py", 1)(Path(u"/testfiles/select")) == 2
        assert select.glob_get_sf(u"/testfiles/select/test.py", 0)(Path(u"/testfiles/select")) is None
        assert select.glob_get_normal_sf(u"/testfiles/select/test.py", 0)(Path(u"/testfiles/select")) is None

    def test_ignore_case(self):
        u"""test_ignore_case - try a few expressions with ignorecase:"""

        sf = self.Select.glob_get_sf(u"ignorecase:testfiles/SeLect/foo/bar", 1)
        assert sf(self.makeext(u"FOO/BAR")) == 1
        assert sf(self.makeext(u"foo/bar")) == 1
        assert sf(self.makeext(u"fOo/BaR")) == 1
        self.assertRaises(FilePrefixError, self.Select.glob_get_sf, u"ignorecase:tesfiles/sect/foo/bar", 1)

    def test_root(self):
        u"""test_root - / may be a counterexample to several of these.."""
        root = Path(u"/")
        select = Select(root)

        self.assertEqual(select.glob_get_sf(u"/", 1)(root), 1)
        self.assertEqual(select.glob_get_sf(u"/foo", 1)(root), 2)
        self.assertEqual(select.glob_get_sf(u"/foo/bar", 1)(root), 2)
        self.assertEqual(select.glob_get_sf(u"/", 0)(root), 0)
        self.assertEqual(select.glob_get_sf(u"/foo", 0)(root), None)

        assert select.glob_get_sf(u"**.py", 1)(root) == 2
        assert select.glob_get_sf(u"**", 1)(root) == 1
        assert select.glob_get_sf(u"ignorecase:/", 1)(root) == 1
        assert select.glob_get_sf(u"**.py", 0)(root) is None
        assert select.glob_get_sf(u"**", 0)(root) == 0
        assert select.glob_get_sf(u"/foo/*", 0)(root) is None

    def test_other_filesystems(self):
        u"""Test to see if --exclude-other-filesystems works correctly"""
        root = Path(u"/")
        select = Select(root)
        sf = select.other_filesystems_get_sf(0)
        assert sf(root) is None
        if os.path.ismount(u"/usr/bin"):
            sfval = 0
        else:
            sfval = None
        assert sf(Path(u"/usr/bin")) == sfval, \
            u"Assumption: /usr/bin is on the same filesystem as /"
        if os.path.ismount(u"/dev"):
            sfval = 0
        else:
            sfval = None
        assert sf(Path(u"/dev")) == sfval, \
            u"Assumption: /dev is on a different filesystem"
        if os.path.ismount(u"/proc"):
            sfval = 0
        else:
            sfval = None
        assert sf(Path(u"/proc")) == sfval, \
            u"Assumption: /proc is on a different filesystem"


class ParseArgsTest(UnitTestCase):
    u"""Test argument parsing"""
    def setUp(self):
        super(ParseArgsTest, self).setUp()
        self.unpack_testfiles()
        self.root = None
        self.expected_restored_tree = [(), (u"1",), (u"1", u"1sub1"), (u"1", u"1sub1", u"1sub1sub1"),
                                       (u"1", u"1sub1", u"1sub1sub1", u"1sub1sub1_file.txt"),
                                       (u"1", u"1sub1", u"1sub1sub3"), (u"1", u"1sub2"), (u"1", u"1sub2", u"1sub2sub1"),
                                       (u"1", u"1sub3"), (u"1", u"1sub3", u"1sub3sub3"), (u"1.py",), (u"2",),
                                       (u"2", u"2sub1"), (u"2", u"2sub1", u"2sub1sub1"),
                                       (u"2", u"2sub1", u"2sub1sub1", u"2sub1sub1_file.txt"),
                                       (u"3",), (u"3", u"3sub2"), (u"3", u"3sub2", u"3sub2sub1"),
                                       (u"3", u"3sub2", u"3sub2sub2"), (u"3", u"3sub2", u"3sub2sub3"), (u"3", u"3sub3"),
                                       (u"3", u"3sub3", u"3sub3sub1"), (u"3", u"3sub3", u"3sub3sub2"),
                                       (u"3", u"3sub3", u"3sub3sub2", u"3sub3sub2_file.txt"),
                                       (u"3", u"3sub3", u"3sub3sub3")]

    def uc_index_from_path(self, path):
        u"""Takes a path type and returns path.index, with each element converted into unicode"""
        uindex = tuple([element.decode(sys.getfilesystemencoding(), u"strict") for element in path.index])
        return uindex

    def ParseTest(self, tuplelist, indicies, filelists=[]):
        u"""No error if running select on tuple goes over indicies"""
        if not self.root:
            self.root = Path(u"testfiles/select")
        self.Select = Select(self.root)
        self.Select.ParseArgs(tuplelist, self.remake_filelists(filelists))
        self.Select.set_iter()

        # Create a list of the paths returned by the select function, converted
        # into path.index styled tuples
        results_as_list = list(Iter.map(self.uc_index_from_path, self.Select))
        # print(results_as_list)
        self.assertEqual(indicies, results_as_list)

    def remake_filelists(self, filelist):
        u"""Turn strings in filelist into fileobjs"""
        new_filelists = []
        for f in filelist:
            if isinstance(f, u"".__class__):
                new_filelists.append(io.StringIO(f))
            else:
                new_filelists.append(f)
        return new_filelists

    def test_parse(self):
        u"""Test just one include, all exclude"""
        self.ParseTest([(u"--include", u"testfiles/select/1/1"),
                        (u"--exclude", u"**")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"1", u"1"),
                        (u"1", u"1", u"2"), (u"1", u"1", u"3")])

    def test_parse2(self):
        u"""Test three level include/exclude"""
        self.ParseTest([(u"--exclude", u"testfiles/select/1/1/1"),
                        (u"--include", u"testfiles/select/1/1"),
                        (u"--exclude", u"testfiles/select/1"),
                        (u"--exclude", u"**")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"1", u"2"),
                        (u"1", u"1", u"3")])

    def test_filelist(self):
        u"""Filelist glob test similar to above testParse2"""
        self.ParseTest([(u"--include-filelist", u"file")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"1", u"2"),
                        (u"1", u"1", u"3")],
                       [u"- testfiles/select/1/1/1\n"
                        u"testfiles/select/1/1\n"
                        u"- testfiles/select/1\n"
                        u"- **"])

    def test_include_filelist_1_trailing_whitespace(self):
        u"""Filelist glob test similar to globbing filelist, but with 1 trailing whitespace on include"""
        self.ParseTest([(u"--include-filelist", u"file")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"1", u"2"),
                        (u"1", u"1", u"3")],
                       [u"- testfiles/select/1/1/1\n"
                        u"testfiles/select/1/1 \n"
                        u"- testfiles/select/1\n"
                        u"- **"])

    def test_include_filelist_2_trailing_whitespaces(self):
        u"""Filelist glob test similar to globbing filelist, but with 2 trailing whitespaces on include"""
        self.ParseTest([(u"--include-filelist", u"file")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"1", u"2"),
                        (u"1", u"1", u"3")],
                       [u"- testfiles/select/1/1/1\n"
                        u"testfiles/select/1/1  \n"
                        u"- testfiles/select/1\n"
                        u"- **"])

    def test_include_filelist_1_leading_whitespace(self):
        u"""Filelist glob test similar to globbing filelist, but with 1 leading whitespace on include"""
        self.ParseTest([(u"--include-filelist", u"file")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"1", u"2"),
                        (u"1", u"1", u"3")],
                       [u"- testfiles/select/1/1/1\n"
                        u" testfiles/select/1/1\n"
                        u"- testfiles/select/1\n"
                        u"- **"])

    def test_include_filelist_2_leading_whitespaces(self):
        u"""Filelist glob test similar to globbing filelist, but with 2 leading whitespaces on include"""
        self.ParseTest([(u"--include-filelist", u"file")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"1", u"2"),
                        (u"1", u"1", u"3")],
                       [u"- testfiles/select/1/1/1\n"
                        u"  testfiles/select/1/1\n"
                        u"- testfiles/select/1\n"
                        u"- **"])

    def test_include_filelist_1_trailing_whitespace_exclude(self):
        u"""Filelist glob test similar to globbing filelist, but with 1 trailing whitespace on exclude"""
        self.ParseTest([(u"--include-filelist", u"file")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"1", u"2"),
                        (u"1", u"1", u"3")],
                       [u"- testfiles/select/1/1/1 \n"
                        u"testfiles/select/1/1\n"
                        u"- testfiles/select/1\n"
                        u"- **"])

    def test_include_filelist_2_trailing_whitespace_exclude(self):
        u"""Filelist glob test similar to globbing filelist, but with 2 trailing whitespaces on exclude"""
        self.ParseTest([(u"--include-filelist", u"file")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"1", u"2"),
                        (u"1", u"1", u"3")],
                       [u"- testfiles/select/1/1/1  \n"
                        u"testfiles/select/1/1\n"
                        u"- testfiles/select/1\n"
                        u"- **"])

    def test_include_filelist_1_leading_whitespace_exclude(self):
        u"""Filelist glob test similar to globbing filelist, but with 1 leading whitespace on exclude"""
        self.ParseTest([(u"--include-filelist", u"file")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"1", u"2"),
                        (u"1", u"1", u"3")],
                       [u" - testfiles/select/1/1/1\n"
                        u"testfiles/select/1/1\n"
                        u"- testfiles/select/1\n"
                        u"- **"])

    def test_include_filelist_2_leading_whitespaces_exclude(self):
        u"""Filelist glob test similar to globbing filelist, but with 2 leading whitespaces on exclude"""
        self.ParseTest([(u"--include-filelist", u"file")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"1", u"2"),
                        (u"1", u"1", u"3")],
                       [u"  - testfiles/select/1/1/1\n"
                        u"testfiles/select/1/1\n"
                        u"- testfiles/select/1\n"
                        u"- **"])

    def test_include_filelist_check_excluded_folder_included_for_contents(self):
        u"""Filelist glob test to check excluded folder is included if contents are"""
        self.ParseTest([(u"--include-filelist", u"file")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"1", u"1"), (u"1", u"1", u"2"),
                        (u"1", u"1", u"3"), (u"1", u"2"), (u"1", u"2", u"1"), (u"1", u"3"), (u"1", u"3", u"1"),
                        (u"1", u"3", u"2"), (u"1", u"3", u"3")],
                       [u"+ testfiles/select/1/2/1\n"
                        u"- testfiles/select/1/2\n"
                        u"testfiles/select/1\n"
                        u"- **"])

    def test_include_filelist_with_unnecessary_quotes(self):
        u"""Filelist glob test similar to globbing filelist, but with quotes around one of the paths."""
        self.ParseTest([(u"--include-filelist", u"file")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"1", u"2"),
                        (u"1", u"1", u"3")],
                       [u"- 'testfiles/select/1/1/1'\n"
                        u"testfiles/select/1/1\n"
                        u"- testfiles/select/1\n"
                        u"- **"])

    def test_include_filelist_with_unnecessary_double_quotes(self):
        u"""Filelist glob test similar to globbing filelist, but with double quotes around one of the paths."""
        self.ParseTest([(u"--include-filelist", u"file")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"1", u"2"),
                        (u"1", u"1", u"3")],
                       [u'- "testfiles/select/1/1/1"\n'
                        u"testfiles/select/1/1\n"
                        u"- testfiles/select/1\n"
                        u"- **"])

    def test_include_filelist_with_full_line_comment(self):
        u"""Filelist glob test similar to globbing filelist, but with a full-line comment."""
        self.ParseTest([(u"--include-filelist", u"file")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"1", u"2"),
                        (u"1", u"1", u"3")],
                       [u"- testfiles/select/1/1/1\n"
                        u"# This is a test\n"
                        u"testfiles/select/1/1\n"
                        u"- testfiles/select/1\n"
                        u"- **"])

    def test_include_filelist_with_blank_line(self):
        u"""Filelist glob test similar to globbing filelist, but with a blank line."""
        self.ParseTest([(u"--include-filelist", u"file")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"1", u"2"),
                        (u"1", u"1", u"3")],
                       [u"- testfiles/select/1/1/1\n"
                        u"\n"
                        u"testfiles/select/1/1\n"
                        u"- testfiles/select/1\n"
                        u"- **"])

    def test_include_filelist_with_blank_line_and_whitespace(self):
        u"""Filelist glob test similar to globbing filelist, but with a blank line and whitespace."""
        self.ParseTest([(u"--include-filelist", u"file")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"1", u"2"),
                        (u"1", u"1", u"3")],
                       [u"- testfiles/select/1/1/1\n"
                        u"  \n"
                        u"testfiles/select/1/1\n"
                        u"- testfiles/select/1\n"
                        u"- **"])

    def test_include_filelist_asterisk(self):
        u"""Filelist glob test with * instead of 'testfiles'"""
        # Thank you to Elifarley Cruz for this test case
        # (https://bugs.launchpad.net/duplicity/+bug/884371).
        self.ParseTest([(u"--include-filelist", u"file")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"1", u"1"),
                        (u"1", u"1", u"2"), (u"1", u"1", u"3")],
                       [u"*/select/1/1\n"
                        u"- **"])

    def test_include_filelist_asterisk_2(self):
        u"""Identical to test_filelist, but with the exclude "select" replaced with '*'"""
        self.ParseTest([(u"--include-filelist", u"file")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"1", u"2"),
                        (u"1", u"1", u"3")],
                       [u"- testfiles/*/1/1/1\n"
                        u"testfiles/select/1/1\n"
                        u"- testfiles/select/1\n"
                        u"- **"])

    def test_include_filelist_asterisk_3(self):
        u"""Identical to test_filelist, but with the auto-include "select" replaced with '*'"""
        # Regression test for Bug #884371 (https://bugs.launchpad.net/duplicity/+bug/884371)
        self.ParseTest([(u"--include-filelist", u"file")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"1", u"2"),
                        (u"1", u"1", u"3")],
                       [u"- testfiles/select/1/1/1\n"
                        u"testfiles/*/1/1\n"
                        u"- testfiles/select/1\n"
                        u"- **"])

    def test_include_filelist_asterisk_4(self):
        u"""Identical to test_filelist, but with a specific include "select" replaced with '*'"""
        # Regression test for Bug #884371 (https://bugs.launchpad.net/duplicity/+bug/884371)
        self.ParseTest([(u"--include-filelist", u"file")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"1", u"2"),
                        (u"1", u"1", u"3")],
                       [u"- testfiles/select/1/1/1\n"
                        u"+ testfiles/*/1/1\n"
                        u"- testfiles/select/1\n"
                        u"- **"])

    def test_include_filelist_asterisk_5(self):
        u"""Identical to test_filelist, but with all 'select's replaced with '*'"""
        # Regression test for Bug #884371 (https://bugs.launchpad.net/duplicity/+bug/884371)
        self.ParseTest([(u"--include-filelist", u"file")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"1", u"2"),
                        (u"1", u"1", u"3")],
                       [u"- testfiles/*/1/1/1\n"
                        u"+ testfiles/*/1/1\n"
                        u"- testfiles/*/1\n"
                        u"- **"])

    def test_include_filelist_asterisk_6(self):
        u"""Identical to test_filelist, but with numerous excluded folders replaced with '*'"""
        self.ParseTest([(u"--include-filelist", u"file")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"1", u"2"),
                        (u"1", u"1", u"3")],
                       [u"- */*/1/1/1\n"
                        u"+ testfiles/select/1/1\n"
                        u"- */*/1\n"
                        u"- **"])

    def test_include_filelist_asterisk_7(self):
        u"""Identical to test_filelist, but with numerous included/excluded folders replaced with '*'"""
        # Regression test for Bug #884371 (https://bugs.launchpad.net/duplicity/+bug/884371)
        self.ParseTest([(u"--include-filelist", u"file")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"1", u"2"),
                        (u"1", u"1", u"3")],
                       [u"- */*/1/1/1\n"
                        u"+ */*/1/1\n"
                        u"- */*/1\n"
                        u"- **"])

    def test_include_filelist_double_asterisk_1(self):
        u"""Identical to test_filelist, but with the exclude "select' replaced with '**'"""
        self.ParseTest([(u"--include-filelist", u"file")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"1", u"2"),
                        (u"1", u"1", u"3")],
                       [u"- testfiles/**/1/1/1\n"
                        u"testfiles/select/1/1\n"
                        u"- testfiles/select/1\n"
                        u"- **"])

    def test_include_filelist_double_asterisk_2(self):
        u"""Identical to test_filelist, but with the include 'select' replaced with '**'"""
        # Regression test for Bug #884371 (https://bugs.launchpad.net/duplicity/+bug/884371)
        self.ParseTest([(u"--include-filelist", u"file")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"1", u"2"),
                        (u"1", u"1", u"3")],
                       [u"- testfiles/select/1/1/1\n"
                        u"**ct/1/1\n"
                        u"- testfiles/select/1\n"
                        u"- **"])

    def test_include_filelist_double_asterisk_3(self):
        u"""Identical to test_filelist, but with the exclude 'testfiles/select' replaced with '**'"""
        self.ParseTest([(u"--include-filelist", u"file")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"1", u"2"),
                        (u"1", u"1", u"3")],
                       [u"- **/1/1/1\n"
                        u"testfiles/select/1/1\n"
                        u"- testfiles/select/1\n"
                        u"- **"])

    def test_include_filelist_double_asterisk_4(self):
        u"""Identical to test_filelist, but with the include 'testfiles/select' replaced with '**'"""
        # Regression test for Bug #884371 (https://bugs.launchpad.net/duplicity/+bug/884371)
        self.ParseTest([(u"--include-filelist", u"file")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"1", u"2"),
                        (u"1", u"1", u"3")],
                       [u"- testfiles/select/1/1/1\n"
                        u"**t/1/1\n"
                        u"- testfiles/select/1\n"
                        u"- **"])

    def test_include_filelist_double_asterisk_5(self):
        u"""Identical to test_filelist, but with all 'testfiles/select's replaced with '**'"""
        # Regression test for Bug #884371 (https://bugs.launchpad.net/duplicity/+bug/884371)
        self.ParseTest([(u"--include-filelist", u"file")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"1", u"2"),
                        (u"1", u"1", u"3")],
                       [u"- **/1/1/1\n"
                        u"**t/1/1\n"
                        u"- **t/1\n"
                        u"- **"])

    def test_include_filelist_trailing_slashes(self):
        u"""Filelist glob test similar to globbing filelist, but with trailing slashes"""
        self.ParseTest([(u"--include-filelist", u"file")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"1", u"2"),
                        (u"1", u"1", u"3")],
                       [u"- testfiles/select/1/1/1/\n"
                        u"testfiles/select/1/1/\n"
                        u"- testfiles/select/1/\n"
                        u"- **"])

    def test_include_filelist_trailing_slashes_and_single_asterisks(self):
        u"""Filelist glob test similar to globbing filelist, but with trailing slashes and single asterisks"""
        # Regression test for Bug #932482 (https://bugs.launchpad.net/duplicity/+bug/932482)
        self.ParseTest([(u"--include-filelist", u"file")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"1", u"2"),
                        (u"1", u"1", u"3")],
                       [u"- */select/1/1/1/\n"
                        u"testfiles/select/1/1/\n"
                        u"- testfiles/*/1/\n"
                        u"- **"])

    def test_include_filelist_trailing_slashes_and_double_asterisks(self):
        u"""Filelist glob test similar to globbing filelist, but with trailing slashes and double asterisks"""
        # Regression test for Bug #932482 (https://bugs.launchpad.net/duplicity/+bug/932482)
        self.ParseTest([(u"--include-filelist", u"file")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"1", u"2"),
                        (u"1", u"1", u"3")],
                       [u"- **/1/1/1/\n"
                        u"testfiles/select/1/1/\n"
                        u"- **t/1/\n"
                        u"- **"])

    def test_filelist_null_separator(self):
        u"""test_filelist, but with null_separator set"""
        self.set_global(u"null_separator", 1)
        self.ParseTest([(u"--include-filelist", u"file")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"1", u"2"),
                        (u"1", u"1", u"3")],
                       [u"\0- testfiles/select/1/1/1\0testfiles/select/1/1\0- testfiles/select/1\0- **\0"])

    def test_exclude_filelist(self):
        u"""Exclude version of test_filelist"""
        self.ParseTest([(u"--exclude-filelist", u"file")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"1", u"2"),
                        (u"1", u"1", u"3")],
                       [u"testfiles/select/1/1/1\n"
                        u"+ testfiles/select/1/1\n"
                        u"testfiles/select/1\n"
                        u"- **"])

    def test_exclude_filelist_asterisk_1(self):
        u"""Exclude version of test_include_filelist_asterisk"""
        self.ParseTest([(u"--exclude-filelist", u"file")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"1", u"1"),
                        (u"1", u"1", u"2"), (u"1", u"1", u"3")],
                       [u"+ */select/1/1\n"
                        u"- **"])

    def test_exclude_filelist_asterisk_2(self):
        u"""Identical to test_exclude_filelist, but with the exclude "select" replaced with '*'"""
        self.ParseTest([(u"--exclude-filelist", u"file")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"1", u"2"),
                        (u"1", u"1", u"3")],
                       [u"testfiles/*/1/1/1\n"
                        u"+ testfiles/select/1/1\n"
                        u"testfiles/select/1\n"
                        u"- **"])

    def test_exclude_filelist_asterisk_3(self):
        u"""Identical to test_exclude_filelist, but with the include "select" replaced with '*'"""
        # Regression test for Bug #884371 (https://bugs.launchpad.net/duplicity/+bug/884371)
        self.ParseTest([(u"--exclude-filelist", u"file")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"1", u"2"),
                        (u"1", u"1", u"3")],
                       [u"testfiles/select/1/1/1\n"
                        u"+ testfiles/*/1/1\n"
                        u"testfiles/select/1\n"
                        u"- **"])

    def test_exclude_filelist_asterisk_4(self):
        u"""Identical to test_exclude_filelist, but with numerous excluded folders replaced with '*'"""
        self.ParseTest([(u"--exclude-filelist", u"file")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"1", u"2"),
                        (u"1", u"1", u"3")],
                       [u"*/select/1/1/1\n"
                        u"+ testfiles/select/1/1\n"
                        u"*/*/1\n"
                        u"- **"])

    def test_exclude_filelist_asterisk_5(self):
        u"""Identical to test_exclude_filelist, but with numerous included/excluded folders replaced with '*'"""
        # Regression test for Bug #884371 (https://bugs.launchpad.net/duplicity/+bug/884371)
        self.ParseTest([(u"--exclude-filelist", u"file")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"1", u"2"),
                        (u"1", u"1", u"3")],
                       [u"*/select/1/1/1\n"
                        u"+ */*/1/1\n"
                        u"*/*/1\n"
                        u"- **"])

    def test_exclude_filelist_double_asterisk(self):
        u"""Identical to test_exclude_filelist, but with all included/excluded folders replaced with '**'"""
        # Regression test for Bug #884371 (https://bugs.launchpad.net/duplicity/+bug/884371)
        self.ParseTest([(u"--exclude-filelist", u"file")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"1", u"2"),
                        (u"1", u"1", u"3")],
                       [u"**/1/1/1\n"
                        u"+ **t/1/1\n"
                        u"**t/1\n"
                        u"- **"])

    def test_exclude_filelist_single_asterisk_at_beginning(self):
        u"""Exclude filelist testing limited functionality of functional test"""
        # Regression test for Bug #884371 (https://bugs.launchpad.net/duplicity/+bug/884371)
        self.root = Path(u"testfiles/select/1")
        self.ParseTest([(u"--exclude-filelist", u"file")],
                       [(), (u"2",), (u"2", u"1")],
                       [u"+ */select/1/2/1\n"
                        u"- testfiles/select/1/2\n"
                        u"- testfiles/*/1/1\n"
                        u"- testfiles/select/1/3"])

    def test_commandline_asterisks_double_both(self):
        u"""Unit test the functional test TestAsterisks.test_commandline_asterisks_double_both"""
        self.root = Path(u"testfiles/select/1")
        self.ParseTest([(u"--include", u"**/1/2/1"),
                        (u"--exclude", u"**t/1/2"),
                        (u"--exclude", u"**t/1/1"),
                        (u"--exclude", u"**t/1/3")],
                       [(), (u"2",), (u"2", u"1")])

    def test_includes_files(self):
        u"""Unit test the functional test test_includes_files"""
        # Test for Bug 1624725
        # https://bugs.launchpad.net/duplicity/+bug/1624725
        self.root = Path(u"testfiles/select2/1/1sub1")
        self.ParseTest([(u"--include", u"testfiles/select2/1/1sub1/1sub1sub1"),
                        (u"--exclude", u"**")],
                       [(), (u"1sub1sub1",), (u"1sub1sub1",
                                             u"1sub1sub1_file.txt")])

    def test_includes_files_trailing_slash(self):
        u"""Unit test the functional test test_includes_files_trailing_slash"""
        # Test for Bug 1624725
        # https://bugs.launchpad.net/duplicity/+bug/1624725
        self.root = Path(u"testfiles/select2/1/1sub1")
        self.ParseTest([(u"--include", u"testfiles/select2/1/1sub1/1sub1sub1/"),
                        (u"--exclude", u"**")],
                       [(), (u"1sub1sub1",), (u"1sub1sub1",
                                             u"1sub1sub1_file.txt")])

    def test_includes_files_trailing_slash_globbing_chars(self):
        u"""Unit test functional test_includes_files_trailing_slash_globbing_chars"""
        # Test for Bug 1624725
        # https://bugs.launchpad.net/duplicity/+bug/1624725
        self.root = Path(u"testfiles/select2/1/1sub1")
        self.ParseTest([(u"--include", u"testfiles/s?lect2/1/1sub1/1sub1sub1/"),
                        (u"--exclude", u"**")],
                       [(), (u"1sub1sub1",), (u"1sub1sub1",
                                             u"1sub1sub1_file.txt")])

    def test_glob(self):
        u"""Test globbing expression"""
        self.ParseTest([(u"--exclude", u"**[3-5]"),
                        (u"--include", u"testfiles/select/1"),
                        (u"--exclude", u"**")],
                       [(), (u"1",), (u"1", u"1"),
                        (u"1", u"1", u"1"), (u"1", u"1", u"2"),
                        (u"1", u"2"), (u"1", u"2", u"1"), (u"1", u"2", u"2")])
        self.ParseTest([(u"--include", u"testfiles/select**/2"),
                        (u"--exclude", u"**")],
                       [(), (u"1",), (u"1", u"1"),
                        (u"1", u"1", u"2"),
                        (u"1", u"2"),
                        (u"1", u"2", u"1"), (u"1", u"2", u"2"), (u"1", u"2", u"3"),
                        (u"1", u"3"),
                        (u"1", u"3", u"2"),
                        (u"2",), (u"2", u"1"),
                        (u"2", u"1", u"1"), (u"2", u"1", u"2"), (u"2", u"1", u"3"),
                        (u"2", u"2"),
                        (u"2", u"2", u"1"), (u"2", u"2", u"2"), (u"2", u"2", u"3"),
                        (u"2", u"3"),
                        (u"2", u"3", u"1"), (u"2", u"3", u"2"), (u"2", u"3", u"3"),
                        (u"3",), (u"3", u"1"),
                        (u"3", u"1", u"2"),
                        (u"3", u"2"),
                        (u"3", u"2", u"1"), (u"3", u"2", u"2"), (u"3", u"2", u"3"),
                        (u"3", u"3"),
                        (u"3", u"3", u"2")])

    def test_filelist2(self):
        u"""Filelist glob test similar to above testGlob"""
        self.ParseTest([(u"--exclude-filelist", u"asoeuth")],
                       [(), (u"1",), (u"1", u"1"),
                        (u"1", u"1", u"1"), (u"1", u"1", u"2"),
                        (u"1", u"2"), (u"1", u"2", u"1"), (u"1", u"2", u"2")],
                       [u"""
**[3-5]
+ testfiles/select/1
**
"""])
        self.ParseTest([(u"--include-filelist", u"file")],
                       [(), (u"1",), (u"1", u"1"),
                        (u"1", u"1", u"2"),
                        (u"1", u"2"),
                        (u"1", u"2", u"1"), (u"1", u"2", u"2"), (u"1", u"2", u"3"),
                        (u"1", u"3"),
                        (u"1", u"3", u"2"),
                        (u"2",), (u"2", u"1"),
                        (u"2", u"1", u"1"), (u"2", u"1", u"2"), (u"2", u"1", u"3"),
                        (u"2", u"2"),
                        (u"2", u"2", u"1"), (u"2", u"2", u"2"), (u"2", u"2", u"3"),
                        (u"2", u"3"),
                        (u"2", u"3", u"1"), (u"2", u"3", u"2"), (u"2", u"3", u"3"),
                        (u"3",), (u"3", u"1"),
                        (u"3", u"1", u"2"),
                        (u"3", u"2"),
                        (u"3", u"2", u"1"), (u"3", u"2", u"2"), (u"3", u"2", u"3"),
                        (u"3", u"3"),
                        (u"3", u"3", u"2")],
                       [u"""
testfiles/select**/2
- **
"""])

    def test_glob2(self):
        u"""Test more globbing functions"""
        self.ParseTest([(u"--include", u"testfiles/select/*foo*/p*"),
                        (u"--exclude", u"**")],
                       [(), (u"efools",), (u"efools", u"ping"),
                        (u"foobar",), (u"foobar", u"pong")])
        self.ParseTest([(u"--exclude", u"testfiles/select/1/1/*"),
                        (u"--exclude", u"testfiles/select/1/2/**"),
                        (u"--exclude", u"testfiles/select/1/3**"),
                        (u"--include", u"testfiles/select/1"),
                        (u"--exclude", u"**")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"2")])

    def test_glob3(self):
        u""" regression test for bug 25230 """
        self.ParseTest([(u"--include", u"testfiles/select/**1"),
                        (u"--include", u"testfiles/select/**2"),
                        (u"--exclude", u"**")],
                       [(), (u"1",), (u"1", u"1"),
                        (u"1", u"1", u"1"), (u"1", u"1", u"2"), (u"1", u"1", u"3"),
                        (u"1", u"2"),
                        (u"1", u"2", u"1"), (u"1", u"2", u"2"), (u"1", u"2", u"3"),
                        (u"1", u"3"),
                        (u"1", u"3", u"1"), (u"1", u"3", u"2"), (u"1", u"3", u"3"),
                        (u"2",), (u"2", u"1"),
                        (u"2", u"1", u"1"), (u"2", u"1", u"2"), (u"2", u"1", u"3"),
                        (u"2", u"2"),
                        (u"2", u"2", u"1"), (u"2", u"2", u"2"), (u"2", u"2", u"3"),
                        (u"2", u"3"),
                        (u"2", u"3", u"1"), (u"2", u"3", u"2"), (u"2", u"3", u"3"),
                        (u"3",), (u"3", u"1"),
                        (u"3", u"1", u"1"), (u"3", u"1", u"2"), (u"3", u"1", u"3"),
                        (u"3", u"2"),
                        (u"3", u"2", u"1"), (u"3", u"2", u"2"), (u"3", u"2", u"3"),
                        (u"3", u"3"),
                        (u"3", u"3", u"1"), (u"3", u"3", u"2")])

    def test_alternate_root(self):
        u"""Test select with different root"""
        self.root = Path(u"testfiles/select/1")
        self.ParseTest([(u"--exclude", u"testfiles/select/1/[23]")],
                       [(), (u"1",), (u"1", u"1"), (u"1", u"2"), (u"1", u"3")])

        self.root = Path(u"/")
        self.ParseTest([(u"--exclude", u"/tmp/*"),
                        (u"--include", u"/tmp"),
                        (u"--exclude", u"/")],
                       [(), (u"tmp",)])

    def test_exclude_after_scan(self):
        u"""Test select with an exclude after a pattern that would return a scan for that file"""
        self.root = Path(u"testfiles/select2/3")
        self.ParseTest([(u"--include", u"testfiles/select2/3/**file.txt"),
                        (u"--exclude", u"testfiles/select2/3/3sub2"),
                        (u"--include", u"testfiles/select2/3/3sub1"),
                        (u"--exclude", u"**")],
                       [(), (u"3sub1",), (u"3sub1", u"3sub1sub1"), (u"3sub1", u"3sub1sub2"), (u"3sub1", u"3sub1sub3"),
                        (u"3sub3",), (u"3sub3", u"3sub3sub2"), (u"3sub3", u"3sub3sub2", u"3sub3sub2_file.txt")])

    def test_include_exclude_basic(self):
        u"""Test functional test test_include_exclude_basic as a unittest"""
        self.root = Path(u"testfiles/select2")
        self.ParseTest([(u"--include", u"testfiles/select2/3/3sub3/3sub3sub2/3sub3sub2_file.txt"),
                        (u"--exclude", u"testfiles/select2/3/3sub3/3sub3sub2"),
                        (u"--include", u"testfiles/select2/3/3sub2/3sub2sub2"),
                        (u"--include", u"testfiles/select2/3/3sub3"),
                        (u"--exclude", u"testfiles/select2/3/3sub1"),
                        (u"--exclude", u"testfiles/select2/2/2sub1/2sub1sub3"),
                        (u"--exclude", u"testfiles/select2/2/2sub1/2sub1sub2"),
                        (u"--include", u"testfiles/select2/2/2sub1"),
                        (u"--exclude", u"testfiles/select2/1/1sub3/1sub3sub2"),
                        (u"--exclude", u"testfiles/select2/1/1sub3/1sub3sub1"),
                        (u"--exclude", u"testfiles/select2/1/1sub2/1sub2sub3"),
                        (u"--include", u"testfiles/select2/1/1sub2/1sub2sub1"),
                        (u"--exclude", u"testfiles/select2/1/1sub1/1sub1sub3/1sub1sub3_file.txt"),
                        (u"--exclude", u"testfiles/select2/1/1sub1/1sub1sub2"),
                        (u"--exclude", u"testfiles/select2/1/1sub2"),
                        (u"--include", u"testfiles/select2/1.py"),
                        (u"--include", u"testfiles/select2/3"),
                        (u"--include", u"testfiles/select2/1"),
                        (u"--exclude", u"testfiles/select2/**")],
                       self.expected_restored_tree)

    def test_globbing_replacement(self):
        u"""Test functional test test_globbing_replacement as a unittest"""
        self.root = Path(u"testfiles/select2")
        self.ParseTest([(u"--include", u"testfiles/select2/**/3sub3sub2/3sub3su?2_file.txt"),
                        (u"--exclude", u"testfiles/select2/*/3s*1"),
                        (u"--exclude", u"testfiles/select2/**/2sub1sub3"),
                        (u"--exclude", u"ignorecase:testfiles/select2/2/2sub1/2Sub1Sub2"),
                        (u"--include", u"ignorecase:testfiles/sel[w,u,e,q]ct2/2/2S?b1"),
                        (u"--exclude", u"testfiles/select2/1/1sub3/1s[w,u,p,q]b3sub2"),
                        (u"--exclude", u"testfiles/select2/1/1sub[1-4]/1sub3sub1"),
                        (u"--include", u"testfiles/select2/1/1sub2/1sub2sub1"),
                        (u"--exclude", u"testfiles/select2/1/1sub1/1sub1sub3/1su?1sub3_file.txt"),
                        (u"--exclude", u"testfiles/select2/1/1*1/1sub1sub2"),
                        (u"--exclude", u"testfiles/select2/1/1sub2"),
                        (u"--include", u"testfiles/select[2-4]/*.py"),
                        (u"--include", u"testfiles/*2/3"),
                        (u"--include", u"**/select2/1"),
                        (u"--exclude", u"testfiles/select2/**")],
                       self.expected_restored_tree)

    def test_unicode_paths_non_globbing(self):
        u"""Test functional test test_unicode_paths_non_globbing as a unittest"""
        self.root = Path(u"testfiles/select-unicode")
        self.ParseTest([(u"--exclude", u"testfiles/select-unicode/прыклад/пример/例/Παράδειγμα/उदाहरण.txt"),
                        (u"--exclude", u"testfiles/select-unicode/прыклад/пример/例/Παράδειγμα/דוגמא.txt"),
                        (u"--exclude", u"testfiles/select-unicode/прыклад/пример/例/მაგალითი/"),
                        (u"--include", u"testfiles/select-unicode/прыклад/пример/例/"),
                        (u"--exclude", u"testfiles/select-unicode/прыклад/пример/"),
                        (u"--include", u"testfiles/select-unicode/прыклад/"),
                        (u"--include", u"testfiles/select-unicode/օրինակ.txt"),
                        (u"--exclude", u"testfiles/select-unicode/**")],
                       [(), (u"прыклад",), (u"прыклад", u"пример"), (u"прыклад", u"пример", u"例"),
                        (u"прыклад", u"пример", u"例", u"Παράδειγμα"),
                        (u"прыклад", u"пример", u"例", u"Παράδειγμα", u"ઉદાહરણ.log"),
                        (u"прыклад", u"উদাহরণ"), (u"օրինակ.txt",)])


class TestGlobGetNormalSf(UnitTestCase):
    u"""Test glob parsing of the test_glob_get_normal_sf function. Indirectly test behaviour of glob_to_re."""

    def glob_tester(self, path, glob_string, include_exclude, root_path):
        u"""Takes a path, glob string and include_exclude value (1 = include, 0 = exclude) and returns the output
        of the selection function.
        None - means the test has nothing to say about the related file
        0 - the file is excluded by the test
        1 - the file is included
        2 - the test says the file (must be directory) should be scanned"""
        self.unpack_testfiles()
        self.root = Path(root_path)
        self.select = Select(self.root)
        selection_function = self.select.glob_get_normal_sf(glob_string, include_exclude)
        path = Path(path)
        return selection_function(path)

    def include_glob_tester(self, path, glob_string, root_path=u"/"):
        return self.glob_tester(path, glob_string, 1, root_path)

    def exclude_glob_tester(self, path, glob_string, root_path=u"/"):
        return self.glob_tester(path, glob_string, 0, root_path)

    def test_glob_get_normal_sf_exclude(self):
        u"""Test simple exclude."""
        self.assertEqual(self.exclude_glob_tester(u"/testfiles/select2/3", u"/testfiles/select2"), 0)
        self.assertEqual(self.exclude_glob_tester(u"/testfiles/.git", u"/testfiles"), 0)

    def test_glob_get_normal_sf_exclude_root(self):
        u"""Test simple exclude with / as the glob."""
        self.assertEqual(self.exclude_glob_tester(u"/.git", u"/"), 0)
        self.assertEqual(self.exclude_glob_tester(u"/testfile", u"/"), 0)

    def test_glob_get_normal_sf_2(self):
        u"""Test same behaviour as the functional test test_globbing_replacement."""
        self.assertEqual(self.include_glob_tester(u"/testfiles/select2/3/3sub3/3sub3sub2/3sub3sub2_file.txt",
                                                  u"/testfiles/select2/**/3sub3sub2/3sub3su?2_file.txt"), 1)
        self.assertEqual(self.include_glob_tester(u"/testfiles/select2/3/3sub1", u"/testfiles/select2/*/3s*1"), 1)
        self.assertEqual(self.include_glob_tester(u"/testfiles/select2/2/2sub1/2sub1sub3",
                                                  u"/testfiles/select2/**/2sub1sub3"), 1)
        self.assertEqual(self.include_glob_tester(u"/testfiles/select2/2/2sub1",
                                                  u"/testfiles/sel[w,u,e,q]ct2/2/2s?b1"), 1)
        self.assertEqual(self.include_glob_tester(u"/testfiles/select2/1/1sub3/1sub3sub2",
                                                  u"/testfiles/select2/1/1sub3/1s[w,u,p,q]b3sub2"), 1)
        self.assertEqual(self.exclude_glob_tester(u"/testfiles/select2/1/1sub3/1sub3sub1",
                                                  u"/testfiles/select2/1/1sub[1-4]/1sub3sub1"), 0)
        self.assertEqual(self.include_glob_tester(u"/testfiles/select2/1/1sub2/1sub2sub1",
                                                  u"/testfiles/select2/*/1sub2/1s[w,u,p,q]b2sub1"), 1)
        self.assertEqual(self.include_glob_tester(u"/testfiles/select2/1/1sub1/1sub1sub3/1sub1sub3_file.txt",
                                                  u"/testfiles/select2/1/1sub1/1sub1sub3/1su?1sub3_file.txt"), 1)
        self.assertEqual(self.exclude_glob_tester(u"/testfiles/select2/1/1sub1/1sub1sub2",
                                                  u"/testfiles/select2/1/1*1/1sub1sub2"), 0)
        self.assertEqual(self.include_glob_tester(u"/testfiles/select2/1/1sub2", u"/testfiles/select2/1/1sub2"), 1)
        self.assertEqual(self.include_glob_tester(u"/testfiles/select2/1.py", u"/testfiles/select[2-4]/*.py"), 1)
        self.assertEqual(self.exclude_glob_tester(u"/testfiles/select2/3", u"/testfiles/*2/3"), 0)
        self.assertEqual(self.include_glob_tester(u"/testfiles/select2/1", u"**/select2/1"), 1)

    def test_glob_get_normal_sf_negative_square_brackets_specified(self):
        u"""Test negative square bracket (specified) [!a,b,c] replacement in get_normal_sf."""
        # As in a normal shell, [!...] expands to any single character but those specified
        self.assertEqual(self.include_glob_tester(u"/test/hello1.txt", u"/test/hello[!2,3,4].txt"), 1)
        self.assertEqual(self.include_glob_tester(u"/test/hello.txt", u"/t[!w,f,h]st/hello.txt"), 1)
        self.assertEqual(self.exclude_glob_tester(u"/long/example/path/hello.txt",
                                                  u"/lon[!w,e,f]/e[!p]ample/path/hello.txt"), 0)
        self.assertEqual(self.include_glob_tester(u"/test/hello1.txt", u"/test/hello[!2,1,3,4].txt"), None)
        self.assertEqual(self.include_glob_tester(u"/test/hello.txt", u"/t[!e,f,h]st/hello.txt"), None)
        self.assertEqual(self.exclude_glob_tester(u"/long/example/path/hello.txt",
                                                  u"/lon[!w,e,g,f]/e[!p,x]ample/path/hello.txt"), None)

    def test_glob_get_normal_sf_negative_square_brackets_range(self):
        u"""Test negative square bracket (range) [!a,b,c] replacement in get_normal_sf."""
        # As in a normal shell, [!1-5] or [!a-f] expands to any single character not in the range specified
        self.assertEqual(self.include_glob_tester(u"/test/hello1.txt", u"/test/hello[!2-4].txt"), 1)
        self.assertEqual(self.include_glob_tester(u"/test/hello.txt", u"/t[!f-h]st/hello.txt"), 1)
        self.assertEqual(self.exclude_glob_tester(u"/long/example/path/hello.txt",
                                                  u"/lon[!w,e,f]/e[!p-s]ample/path/hello.txt"), 0)
        self.assertEqual(self.include_glob_tester(u"/test/hello1.txt", u"/test/hello[!1-4].txt"), None)
        self.assertEqual(self.include_glob_tester(u"/test/hello.txt", u"/t[!b-h]st/hello.txt"), None)
        self.assertEqual(self.exclude_glob_tester(u"/long/example/path/hello.txt",
                                                  u"/lon[!f-p]/e[!p]ample/path/hello.txt"), None)

    def test_glob_get_normal_sf_2_ignorecase(self):
        u"""Test same behaviour as the functional test test_globbing_replacement, ignorecase tests."""
        self.assertEqual(self.include_glob_tester(u"testfiles/select2/2/2sub1",
                                                  u"ignorecase:testfiles/sel[w,u,e,q]ct2/2/2S?b1",
                                                  u"testfiles/select2"), 1)
        self.assertEqual(self.include_glob_tester(u"testfiles/select2/2/2sub1/2sub1sub2",
                                                  u"ignorecase:testfiles/select2/2/2sub1/2Sub1Sub2",
                                                  u"testfiles/select2"), 1)

    def test_glob_get_normal_sf_3_double_asterisks_dirs_to_scan(self):
        u"""Test double asterisk (**) replacement in glob_get_normal_sf with directories that should be scanned"""
        # The new special pattern, **, expands to any string of characters whether or not it contains "/".
        self.assertEqual(self.include_glob_tester(u"/long/example/path", u"/**/hello.txt"), 2)

    def test_glob_get_normal_sf_3_ignorecase(self):
        u"""Test ignorecase in glob_get_normal_sf"""
        # If the pattern starts with "ignorecase:" (case insensitive), then this prefix will be removed and any
        # character in the string can be replaced with an upper- or lowercase version of itself.
        self.assertEqual(self.include_glob_tester(u"testfiles/select2/2", u"ignorecase:testfiles/select2/2",
                                                  u"testfiles/select2"), 1)
        self.assertEqual(self.include_glob_tester(u"testfiles/select2/2", u"ignorecase:testFiles/Select2/2",
                                                  u"testfiles/select2"), 1)
        self.assertEqual(self.include_glob_tester(u"tEstfiles/seLect2/2", u"ignorecase:testFiles/Select2/2",
                                                  u"testfiles/select2"), 1)
        self.assertEqual(self.include_glob_tester(u"TEstfiles/SeLect2/2", u"ignorecase:t?stFiles/S*ect2/2",
                                                  u"testfiles/select2"), 1)
        self.assertEqual(self.include_glob_tester(u"TEstfiles/SeLect2/2", u"ignorecase:t?stFil**ect2/2",
                                                  u"testfiles/select2"), 1)
        self.assertEqual(self.exclude_glob_tester(u"TEstfiles/SeLect2/2", u"ignorecase:t?stFiles/S*ect2/2",
                                                  u"testfiles/select2"), 0)
        self.assertEqual(self.exclude_glob_tester(u"TEstFiles/SeLect2/2", u"ignorecase:t?stFile**ect2/2",
                                                  u"testfiles/select2"), 0)

    def test_glob_dirs_to_scan(self):
        u"""Test parent directories are marked as needing to be scanned"""
        with patch(u"duplicity.path.Path.isdir") as mock_isdir:
            mock_isdir.return_value = True
            self.assertEqual(
                self.glob_tester(u"parent", u"parent/hello.txt", 1, u"parent"), 2)

    def test_glob_dirs_to_scan_glob(self):
        u"""Test parent directories are marked as needing to be scanned - globs"""
        with patch(u"duplicity.path.Path.isdir") as mock_isdir:
            mock_isdir.return_value = True
            self.assertEqual(
                self.glob_tester(u"testfiles/select/1", u"*/select/1/1", 1,
                                 u"testfiles/select"), 2)
            self.assertEqual(
                self.glob_tester(u"testfiles/select/1/2",
                                 u"*/select/1/2/1", 1, u"testfiles/select"), 2)
            self.assertEqual(
                self.glob_tester(u"parent", u"parent/hel?o.txt", 1, u"parent"), 2)
            self.assertEqual(
                self.glob_tester(u"test/parent/folder",
                                 u"test/par*t/folder/hello.txt", 1, u"test"), 2)
            self.assertEqual(
                self.glob_tester(u"testfiles/select/1/1",
                                 u"**/1/2/1", 1, u"testfiles"), 2)
            self.assertEqual(
                self.glob_tester(u"testfiles/select2/3/3sub2",
                                 u"testfiles/select2/3/**file.txt", 1,
                                 u"testfiles"), 2)
            self.assertEqual(
                self.glob_tester(u"testfiles/select/1/2",
                                 u"*/select/1/2/1", 1, u"testfiles"), 2)
            self.assertEqual(
                self.glob_tester(u"testfiles/select/1",
                                 u"testfiles/select**/2", 1, u"testfiles"), 2)
            self.assertEqual(
                self.glob_tester(u"testfiles/select/efools",
                                 u"testfiles/select/*foo*/p*", 1,
                                 u"testfiles"), 2)
            self.assertEqual(
                self.glob_tester(u"testfiles/select/3",
                                 u"testfiles/select/**2", 1, u"testfiles"), 2)
            self.assertEqual(
                self.glob_tester(u"testfiles/select2/1/1sub1/1sub1sub2",
                                 u"testfiles/select2/**/3sub3sub2/3sub3su?2_file.txt",
                                 1, u"testfiles"), 2)
            self.assertEqual(
                self.glob_tester(u"testfiles/select/1",
                                 u"*/select/1/1", 1, u"testfiles"), 2)

if __name__ == u"__main__":
    unittest.main()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

from __future__ import print_function
from builtins import range
from future import standard_library
standard_library.install_aliases()

import os
import unittest

from duplicity import path
from . import CmdError, FunctionalTestCase


class ReplicateTest(FunctionalTestCase):
    u"""
    Test backup/replicate/restore using duplicity binary
    """

    def runtest(self, dirlist, backup_options=[], replicate_options=[], restore_options=[]):
        # Back up directories to local backend
        current_time = 100000
        self.backup(u"full", dirlist[0], current_time=current_time,
                    options=backup_options)
        for new_dir in dirlist[1:]:
            current_time += 100000
            self.backup(u"inc", new_dir, current_time=current_time,
                        options=backup_options)

        # Replicate to other backend
        source_url = self.backend_url
        target_url = u"file://testfiles/replicate_out"
        self.run_duplicity(options=[u"replicate"] +
                           replicate_options + [source_url, target_url])

        self.backend_url = target_url

        # Restore each and compare them
        for i in range(len(dirlist)):
            dirname = dirlist[i]
            current_time = 100000 * (i + 1)
            self.restore(time=current_time, options=restore_options)
            self.check_same(dirname, u"testfiles/restore_out")
            self.verify(dirname,
                        time=current_time, options=restore_options)

    def check_same(self, filename1, filename2):
        u"""Verify two filenames are the same"""
        path1, path2 = path.Path(filename1), path.Path(filename2)
        assert path1.compare_recursive(path2, verbose=1)

    def test_replicate(self):
        u"""Test replication"""
        self.runtest([u"testfiles/dir1", u"testfiles/dir2"])

    def test_replicate_noencryption(self):
        u"""Test replication with decryption"""
        self.runtest([u"testfiles/dir1", u"testfiles/dir2"],
                     replicate_options=[u"--no-encryption"])

    def test_replicate_asym(self):
        u"""Test replication with reencryption"""
        asym_options = [u"--encrypt-key", self.encrypt_key1]
        self.runtest([u"testfiles/dir1", u"testfiles/dir2"],
                     replicate_options=asym_options, restore_options=asym_options)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
# Copyright 2008 Michael Terry <mike@mterry.name>
# Copyright 2011 Canonical Ltd
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

u"""Log various messages depending on verbosity level"""
from __future__ import division

from builtins import str
from builtins import object
import os
import sys
import logging
import datetime

MIN = 0
ERROR = 0
WARNING = 2
NOTICE = 3
INFO = 5
DEBUG = 9
MAX = 9

_logger = None


def DupToLoggerLevel(verb):
    u"""Convert duplicity level to the logging module's system, where higher is
       more severe"""
    return MAX - verb + 1


def LoggerToDupLevel(verb):
    u"""Convert logging module level to duplicity's system, where lowere is
       more severe"""
    return DupToLoggerLevel(verb)


def LevelName(level):
    if level >= 9:
        return u"DEBUG"
    elif level >= 5:
        return u"INFO"
    elif level >= 3:
        return u"NOTICE"
    elif level >= 1:
        return u"WARNING"
    else:
        return u"ERROR"


def Log(s, verb_level, code=1, extra=None, force_print=False):
    u"""Write s to stderr if verbosity level low enough"""
    global _logger
    if extra:
        controlLine = u'%d %s' % (code, extra)
    else:
        controlLine = u'%d' % (code)
    if not s:
        s = u''  # If None is passed, standard logging would render it as 'None'

    if force_print:
        initial_level = _logger.getEffectiveLevel()
        _logger.setLevel(DupToLoggerLevel(MAX))

    # If all the backends kindly gave us unicode, we could enable this next
    # assert line.  As it is, we'll attempt to convert s to unicode if we
    # are handed bytes.  One day we should update the backends.
    # assert isinstance(s, unicode)
    if not isinstance(s, str):
        s = s.decode(u"utf8", u"replace")

    _logger.log(DupToLoggerLevel(verb_level), s,
                extra={u'levelName': LevelName(verb_level),
                       u'controlLine': controlLine})

    if force_print:
        _logger.setLevel(initial_level)


def Debug(s):
    u"""Shortcut used for debug message (verbosity 9)."""
    Log(s, DEBUG)


class InfoCode(object):
    u"""Enumeration class to hold info code values.
       These values should never change, as frontends rely upon them.
       Don't use 0 or negative numbers."""
    generic = 1
    progress = 2
    collection_status = 3
    diff_file_new = 4
    diff_file_changed = 5
    diff_file_deleted = 6
    patch_file_writing = 7
    patch_file_patching = 8
    # file_list = 9 # 9 isn't used anymore.  It corresponds to an older syntax for listing files
    file_list = 10
    synchronous_upload_begin = 11
    asynchronous_upload_begin = 12
    synchronous_upload_done = 13
    asynchronous_upload_done = 14
    skipping_socket = 15
    upload_progress = 16


def Info(s, code=InfoCode.generic, extra=None):
    u"""Shortcut used for info messages (verbosity 5)."""
    Log(s, INFO, code, extra)


def Progress(s, current, total=None):
    u"""Shortcut used for progress messages (verbosity 5)."""
    if total:
        controlLine = u'%d %d' % (current, total)
    else:
        controlLine = u'%d' % current
    Log(s, INFO, InfoCode.progress, controlLine)


def _ElapsedSecs2Str(secs):
    tdelta = datetime.timedelta(seconds=secs)
    hours, rem = divmod(tdelta.seconds, 3600)
    minutes, seconds = divmod(rem, 60)
    fmt = u""
    if tdelta.days > 0:
        fmt = u"%dd," % (tdelta.days)
    fmt = u"%s%02d:%02d:%02d" % (fmt, hours, minutes, seconds)
    return fmt


def _RemainingSecs2Str(secs):
    tdelta = datetime.timedelta(seconds=secs)
    hours, rem = divmod(tdelta.seconds, 3600)
    minutes, seconds = divmod(rem, 60)
    fmt = u""
    if tdelta.days > 0:
        fmt = u"%dd" % (tdelta.days)
        if hours > 0:
            fmt = u"%s %dh" % (fmt, hours)
        if minutes > 0:
            fmt = u"%s %dmin" % (fmt, minutes)
    elif hours > 0:
        fmt = u"%dh" % hours
        if minutes > 0:
            fmt = u"%s %dmin" % (fmt, minutes)
    elif minutes > 5:
        fmt = u"%dmin" % minutes
    elif minutes > 0:
        fmt = u"%dmin" % minutes
        if seconds >= 30:
            fmt = u"%s 30sec" % fmt
    elif seconds > 45:
        fmt = u"< 1min"
    elif seconds > 30:
        fmt = u"< 45sec"
    elif seconds > 15:
        fmt = u"< 30sec"
    else:
        fmt = u"%dsec" % seconds
    return fmt


def TransferProgress(progress, eta, changed_bytes, elapsed, speed, stalled):
    u"""Shortcut used for upload progress messages (verbosity 5)."""
    dots = int(0.4 * progress)  # int(40.0 * progress / 100.0) -- for 40 chars
    data_amount = float(changed_bytes) / 1024.0
    data_scale = u"KB"
    if data_amount > 1000.0:
        data_amount /= 1024.0
        data_scale = u"MB"
    if data_amount > 1000.0:
        data_amount /= 1024.0
        data_scale = u"GB"
    if stalled:
        eta_str = u"Stalled!"
        speed_amount = 0
        speed_scale = u"B"
    else:
        eta_str = _RemainingSecs2Str(eta)
        speed_amount = float(speed) / 1024.0
        speed_scale = u"KB"
        if speed_amount > 1000.0:
            speed_amount /= 1024.0
            speed_scale = u"MB"
        if speed_amount > 1000.0:
            speed_amount /= 1024.0
            speed_scale = u"GB"
    s = u"%.1f%s %s [%.1f%s/s] [%s>%s] %d%% ETA %s" % (data_amount, data_scale,
                                                       _ElapsedSecs2Str(elapsed),
                                                       speed_amount, speed_scale,
                                                       u'=' * dots, u' ' * (40 - dots),
                                                       progress,
                                                       eta_str
                                                       )

    controlLine = u"%d %d %d %d %d %d" % (changed_bytes, elapsed, progress, eta, speed, stalled)
    Log(s, NOTICE, InfoCode.upload_progress, controlLine)


def PrintCollectionStatus(col_stats, force_print=False):
    u"""Prints a collection status to the log"""
    Log(str(col_stats), 8, InfoCode.collection_status,
        u'\n' + u'\n'.join(col_stats.to_log_info()), force_print)


def PrintCollectionFileChangedStatus(col_stats, filepath, force_print=False):
    u"""Prints a collection status to the log"""
    Log(str(col_stats.get_file_changed_record(filepath)), 8, InfoCode.collection_status, None, force_print)


def Notice(s):
    u"""Shortcut used for notice messages (verbosity 3, the default)."""
    Log(s, NOTICE)


class WarningCode(object):
    u"""Enumeration class to hold warning code values.
       These values should never change, as frontends rely upon them.
       Don't use 0 or negative numbers."""
    generic = 1
    orphaned_sig = 2
    unnecessary_sig = 3
    unmatched_sig = 4
    incomplete_backup = 5
    orphaned_backup = 6
    ftp_ncftp_v320 = 7  # moved from error
    cannot_iterate = 8
    cannot_stat = 9
    cannot_read = 10
    no_sig_for_time = 11
    cannot_process = 12
    process_skipped = 13


def Warn(s, code=WarningCode.generic, extra=None):
    u"""Shortcut used for warning messages (verbosity 2)"""
    Log(s, WARNING, code, extra)


class ErrorCode(object):
    u"""Enumeration class to hold error code values.
       These values should never change, as frontends rely upon them.
       Don't use 0 or negative numbers.  This code is returned by duplicity
       to indicate which error occurred via both exit code and log."""
    generic = 1  # Don't use if possible, please create a new code and use it
    command_line = 2
    hostname_mismatch = 3
    no_manifests = 4
    mismatched_manifests = 5
    unreadable_manifests = 6
    cant_open_filelist = 7
    bad_url = 8
    bad_archive_dir = 9
    bad_sign_key = 10
    restore_dir_exists = 11
    verify_dir_doesnt_exist = 12
    backup_dir_doesnt_exist = 13
    file_prefix_error = 14
    globbing_error = 15
    redundant_inclusion = 16
    inc_without_sigs = 17
    no_sigs = 18
    restore_dir_not_found = 19
    no_restore_files = 20
    mismatched_hash = 21
    unsigned_volume = 22
    user_error = 23
    boto_old_style = 24
    boto_lib_too_old = 25
    boto_calling_format = 26
    ftp_ncftp_missing = 27
    ftp_ncftp_too_old = 28
    # ftp_ncftp_v320 = 29 # moved to warning
    exception = 30
    gpg_failed = 31
    s3_bucket_not_style = 32
    not_implemented = 33
    get_freespace_failed = 34
    not_enough_freespace = 35
    get_ulimit_failed = 36
    maxopen_too_low = 37
    connection_failed = 38
    restart_file_not_found = 39
    gio_not_available = 40
    source_dir_mismatch = 42  # 41 is reserved for par2
    ftps_lftp_missing = 43
    volume_wrong_size = 44
    enryption_mismatch = 45
    pythonoptimize_set = 46

    dpbx_nologin = 47

    bad_request = 48
    s3_kms_no_id = 49

    # 50->69 reserved for backend errors
    backend_error = 50
    backend_permission_denied = 51
    backend_not_found = 52
    backend_no_space = 53
    backend_command_error = 54
    backend_code_error = 55

    # Reserve 126 because it is used as an error code for pkexec
    # Reserve 127 because it is used as an error code for pkexec
    # Reserve 255 because it is used as an error code for gksu


def Error(s, code=ErrorCode.generic, extra=None):
    u"""Write error message"""
    Log(s, ERROR, code, extra)


def FatalError(s, code=ErrorCode.generic, extra=None):
    u"""Write fatal error message and exit"""
    Log(s, ERROR, code, extra)
    shutdown()
    sys.exit(code)


class OutFilter(logging.Filter):
    u"""Filter that only allows warning or less important messages"""
    def filter(self, record):
        return record.msg and record.levelno <= DupToLoggerLevel(WARNING)


class ErrFilter(logging.Filter):
    u"""Filter that only allows messages more important than warnings"""
    def filter(self, record):
        return record.msg and record.levelno > DupToLoggerLevel(WARNING)


def setup():
    u"""Initialize logging"""
    global _logger
    if _logger:
        return

    _logger = logging.getLogger(u"duplicity")

    # Default verbosity allows notices and above
    setverbosity(NOTICE)

    # stdout and stderr are for different logging levels
    outHandler = logging.StreamHandler(sys.stdout)
    outHandler.addFilter(OutFilter())
    _logger.addHandler(outHandler)

    errHandler = logging.StreamHandler(sys.stderr)
    errHandler.addFilter(ErrFilter())
    _logger.addHandler(errHandler)


class MachineFormatter(logging.Formatter):
    u"""Formatter that creates messages in a syntax easily consumable by other
       processes."""
    def __init__(self):
        # 'message' will be appended by format()
        # Note that we use our own, custom-created 'levelName' instead of the
        # standard 'levelname'.  This is because the standard 'levelname' can
        # be adjusted by any library anywhere in our stack without us knowing.
        # But we control 'levelName'.
        logging.Formatter.__init__(self, u"%(levelName)s %(controlLine)s")

    def format(self, record):
        s = logging.Formatter.format(self, record)

        # Add user-text hint of 'message' back in, with each line prefixed by a
        # dot, so consumers know it's not part of 'controlLine'
        if record.message:
            s += (u'\n' + record.message).replace(u'\n', u'\n. ')

        # Add a newline so consumers know the message is over.
        return s + u'\n'


class MachineFilter(logging.Filter):
    u"""Filter that only allows levels that are consumable by other processes."""
    def filter(self, record):
        # We only want to allow records that have our custom level names
        return hasattr(record, u'levelName')


def add_fd(fd):
    u"""Add stream to which to write machine-readable logging"""
    global _logger
    handler = logging.StreamHandler(os.fdopen(fd, u'w'))
    handler.setFormatter(MachineFormatter())
    handler.addFilter(MachineFilter())
    _logger.addHandler(handler)


def add_file(filename):
    u"""Add file to which to write machine-readable logging"""
    global _logger
    handler = logging.FileHandler(filename, encoding=u'utf8')
    handler.setFormatter(MachineFormatter())
    handler.addFilter(MachineFilter())
    _logger.addHandler(handler)


def setverbosity(verb):
    u"""Set the verbosity level"""
    global _logger
    _logger.setLevel(DupToLoggerLevel(verb))


def getverbosity():
    u"""Get the verbosity level"""
    global _logger
    return LoggerToDupLevel(_logger.getEffectiveLevel())


def shutdown():
    u"""Cleanup and flush loggers"""
    logging.shutdown()



# -*- coding: utf-8 -*-
# Copyright (c) 2014 Gu1
# Licensed under the MIT license

from __future__ import print_function
from future import standard_library
standard_library.install_aliases()
from builtins import str

import configparser
import os
import re
import time
import urllib.parse  # pylint: disable=import-error

from requests.compat import quote, quote_plus
import requests


OAUTH_ENDPOINT = u"https://api.hubic.com/oauth/"
API_ENDPOINT = u"https://api.hubic.com/1.0/"
TOKENS_FILE = os.path.expanduser(u"~/.hubic_tokens")


class BearerTokenAuth(requests.auth.AuthBase):
    def __init__(self, token):
        self.token = token

    def __call__(self, req):
        req.headers[u'Authorization'] = u'Bearer ' + self.token
        return req


class HubicIdentity(BaseIdentity):
    def __init__(self):
        try:
            from pyrax.base_identity import BaseIdentity, Service
            import pyrax
            import pyrax.exceptions as exc
        except ImportError as e:
            raise BackendException(u"""\
Hubic backend requires the pyrax library available from Rackspace.
Exception: %s""" % str(e))

    def _get_auth_endpoint(self):
        return u""

    def set_credentials(self, email, password, client_id,
                        client_secret, redirect_uri,
                        authenticate=False):
        u"""Sets the username and password directly."""
        self._email = email
        self._password = password
        self._client_id = client_id
        self.tenant_id = client_id
        self._client_secret = client_secret
        self._redirect_uri = redirect_uri
        if authenticate:
            self.authenticate()

    def _read_credential_file(self, cfg):
        u"""
        Parses the credential file with Rackspace-specific labels.
        """
        self._email = cfg.get(u"hubic", u"email")
        self._password = cfg.get(u"hubic", u"password")
        self._client_id = cfg.get(u"hubic", u"client_id")
        self.tenant_id = self._client_id
        self._client_secret = cfg.get(u"hubic", u"client_secret")
        self._redirect_uri = cfg.get(u"hubic", u"redirect_uri")

    def _parse_error(self, resp):
        if u'location' not in resp.headers:
            return None
        query = urllib.parse.urlsplit(resp.headers[u'location']).query
        qs = dict(urllib.parse.parse_qsl(query))
        return {u'error': qs[u'error'], u'error_description': qs[u'error_description']}

    def _get_access_token(self, code):
        r = requests.post(
            OAUTH_ENDPOINT + u'token/',
            data={
                u'code': code,
                u'redirect_uri': self._redirect_uri,
                u'grant_type': u'authorization_code',
            },
            auth=(self._client_id, self._client_secret)
        )
        if r.status_code != 200:
            try:
                err = r.json()
                err[u'code'] = r.status_code
            except:
                err = {}

            raise exc.AuthenticationFailed(u"Unable to get oauth access token, "
                                           u"wrong client_id or client_secret ? (%s)" %
                                           str(err))

        oauth_token = r.json()

        config = configparser.ConfigParser()
        config.read(TOKENS_FILE)

        if not config.has_section(u"hubic"):
            config.add_section(u"hubic")

        if oauth_token[u'access_token'] is not None:
            config.set(u"hubic", u"access_token", oauth_token[u'access_token'])
            with open(TOKENS_FILE, u'wb') as configfile:
                config.write(configfile)
        else:
            raise exc.AuthenticationFailed(
                u"Unable to get oauth access token, wrong client_id or client_secret ? (%s)" %
                str(err))

        if oauth_token[u'refresh_token'] is not None:
            config.set(u"hubic", u"refresh_token", oauth_token[u'refresh_token'])
            with open(TOKENS_FILE, u'wb') as configfile:
                config.write(configfile)
        else:
            raise exc.AuthenticationFailed(u"Unable to get the refresh token.")

        # removing username and password from .hubic_tokens
        if config.has_option(u"hubic", u"email"):
            config.remove_option(u"hubic", u"email")
            with open(TOKENS_FILE, u'wb') as configfile:
                config.write(configfile)
            print(u"username has been removed from the .hubic_tokens file sent to the CE.")
        if config.has_option(u"hubic", u"password"):
            config.remove_option(u"hubic", u"password")
            with open(TOKENS_FILE, u'wb') as configfile:
                config.write(configfile)
            print(u"password has been removed from the .hubic_tokens file sent to the CE.")

        return oauth_token

    def _refresh_access_token(self):

        config = configparser.ConfigParser()
        config.read(TOKENS_FILE)
        refresh_token = config.get(u"hubic", u"refresh_token")

        if refresh_token is None:
            raise exc.AuthenticationFailed(u"refresh_token is null. Not acquiered before ?")

        success = False
        max_retries = 20
        retries = 0
        sleep_time = 30
        max_sleep_time = 3600

        while retries < max_retries and not success:
            r = requests.post(
                OAUTH_ENDPOINT + u'token/',
                data={
                    u'refresh_token': refresh_token,
                    u'grant_type': u'refresh_token',
                },
                auth=(self._client_id, self._client_secret)
            )
            if r.status_code != 200:
                if r.status_code == 509:
                    print(u"status_code 509: attempt #", retries, u" failed")
                    retries += 1
                    time.sleep(sleep_time)
                    sleep_time = sleep_time * 2
                    if sleep_time > max_sleep_time:
                        sleep_time = max_sleep_time
                else:
                    try:
                        err = r.json()
                        err[u'code'] = r.status_code
                    except:
                        err = {}

                    raise exc.AuthenticationFailed(
                        u"Unable to get oauth access token, wrong client_id or client_secret ? (%s)" %
                        str(err))
            else:
                success = True

        if not success:
            raise exc.AuthenticationFailed(
                u"All the attempts failed to get the refresh token: "
                u"status_code = 509: Bandwidth Limit Exceeded")

        oauth_token = r.json()

        if oauth_token[u'access_token'] is not None:
            return oauth_token
        else:
            raise exc.AuthenticationFailed(u"Unable to get oauth access token from json")

    def authenticate(self):
        config = configparser.ConfigParser()
        config.read(TOKENS_FILE)

        if config.has_option(u"hubic", u"refresh_token"):
            oauth_token = self._refresh_access_token()
        else:
            r = requests.get(
                OAUTH_ENDPOINT + u'auth/?client_id={0}&redirect_uri={1}'
                u'&scope=credentials.r,account.r&response_type=code&state={2}'.format(
                    quote(self._client_id),
                    quote_plus(self._redirect_uri),
                    pyrax.utils.random_ascii()  # csrf ? wut ?..
                ),
                allow_redirects=False
            )
            if r.status_code != 200:
                raise exc.AuthenticationFailed(u"Incorrect/unauthorized "
                                               u"client_id (%s)" % str(self._parse_error(r)))

            try:
                from lxml import html as lxml_html
            except ImportError:
                lxml_html = None

            if lxml_html:
                oauth = lxml_html.document_fromstring(r.content).xpath(u'//input[@name="oauth"]')
                oauth = oauth[0].value if oauth else None
            else:
                oauth = re.search(
                    r'<input\s+[^>]*name=[\'"]?oauth[\'"]?\s+[^>]*value=[\'"]?(\d+)[\'"]?>',
                    r.content)
                oauth = oauth.group(1) if oauth else None

            if not oauth:
                raise exc.AuthenticationFailed(u"Unable to get oauth_id from authorization page")

            if self._email is None or self._password is None:
                raise exc.AuthenticationFailed(u"Cannot retrieve email and/or password. "
                                               u"Please run expresslane-hubic-setup.sh")

            r = requests.post(
                OAUTH_ENDPOINT + u'auth/',
                data={
                    u'action': u'accepted',
                    u'oauth': oauth,
                    u'login': self._email,
                    u'user_pwd': self._password,
                    u'account': u'r',
                    u'credentials': u'r',

                },
                allow_redirects=False
            )

            try:
                query = urllib.parse.urlsplit(r.headers[u'location']).query
                code = dict(urllib.parse.parse_qsl(query))[u'code']
            except:
                raise exc.AuthenticationFailed(u"Unable to authorize client_id, "
                                               u"invalid login/password ?")

            oauth_token = self._get_access_token(code)

        if oauth_token[u'token_type'].lower() != u'bearer':
            raise exc.AuthenticationFailed(u"Unsupported access token type")

        r = requests.get(
            API_ENDPOINT + u'account/credentials',
            auth=BearerTokenAuth(oauth_token[u'access_token']),
        )

        swift_token = r.json()
        self.authenticated = True
        self.token = swift_token[u'token']
        self.expires = swift_token[u'expires']
        self.services[u'object_store'] = Service(self, {
            u'name': u'HubiC',
            u'type': u'cloudfiles',
            u'endpoints': [
                {u'public_url': swift_token[u'endpoint']}
            ]
        })
        self.username = self.password = None



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2013 J.P. Krauss <jkrauss@asymworks.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

from builtins import str
import os

from duplicity import log
from duplicity import util
from duplicity.errors import BackendException
import duplicity.backend

from ._cf_pyrax import PyraxBackend


class HubicBackend(PyraxBackend):
    u"""
    Backend for Hubic using Pyrax
    """
    def __init__(self, parsed_url):
        duplicity.backend.Backend.__init__(self, parsed_url)

        try:
            import pyrax
        except ImportError as e:
            raise BackendException(u"""\
Hubic backend requires the pyrax library available from Rackspace.
Exception: %s""" % str(e))

        # Inform Pyrax that we're talking to Hubic
        pyrax.set_setting(u"identity_type", u"duplicity.backends.pyrax_identity.hubic.HubicIdentity")

        CREDENTIALS_FILE = os.path.expanduser(u"~/.hubic_credentials")
        if os.path.exists(CREDENTIALS_FILE):
            try:
                pyrax.set_credential_file(CREDENTIALS_FILE)
            except Exception as e:
                log.FatalError(u"Connection failed, please check your credentials: %s %s"
                               % (e.__class__.__name__, util.uexc(e)),
                               log.ErrorCode.connection_failed)

        else:
            raise BackendException(u"No ~/.hubic_credentials file found.")

        container = parsed_url.path.lstrip(u'/')

        self.client_exc = pyrax.exceptions.ClientException
        self.nso_exc = pyrax.exceptions.NoSuchObject
        self.container = pyrax.cloudfiles.create_container(container)


duplicity.backend.register_backend(u"cf+hubic", HubicBackend)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

from __future__ import print_function
from future import standard_library
standard_library.install_aliases()
from builtins import next
from builtins import range

import unittest
import pickle
import sys
from functools import reduce

from duplicity.lazy import *  # @UnusedWildImport
from . import UnitTestCase


class Iterators(UnitTestCase):
    one_to_100 = lambda s: iter(list(range(1, 101)))
    evens = lambda s: iter(list(range(2, 101, 2)))
    odds = lambda s: iter(list(range(1, 100, 2)))
    empty = lambda s: iter([])

    def __init__(self, *args):
        super(Iterators, self).__init__(*args)
        self.falseerror = self.falseerror_maker()
        self.trueerror = self.trueerror_maker()
        self.emptygen = self.emptygen_maker()
        self.typeerror = self.typeerror_maker()
        self.nameerror = self.nameerror_maker()

    def falseerror_maker(self):
        yield None
        yield 0
        yield []
        raise Exception

    def trueerror_maker(self):
        yield 1
        yield u"hello"
        yield (2, 3)
        raise Exception

    def nameerror_maker(self):
        if 0:
            yield 1
        raise NameError

    def typeerror_maker(self):
        yield 1
        yield 2
        raise TypeError

    def alwayserror(self, x):
        raise Exception

    def emptygen_maker(self):
        if 0:
            yield 1


class IterEqualTestCase(Iterators):
    u"""Tests for iter_equal function"""
    def testEmpty(self):
        u"""Empty iterators should be equal"""
        assert Iter.equal(self.empty(), iter([]))

    def testNormal(self):
        u"""See if normal iterators are equal"""
        assert Iter.equal(iter((1, 2, 3)), iter((1, 2, 3)))
        assert Iter.equal(self.odds(), iter(list(range(1, 100, 2))))
        assert Iter.equal(iter((1, 2, 3)), iter(list(range(1, 4))))

    def testNormalInequality(self):
        u"""See if normal unequals work"""
        assert not Iter.equal(iter((1, 2, 3)), iter((1, 2, 4)))
        assert not Iter.equal(self.odds(), iter([u"hello", u"there"]))

    def testGenerators(self):
        u"""equals works for generators"""
        def f():
            yield 1
            yield u"hello"

        def g():
            yield 1
            yield u"hello"

        assert Iter.equal(f(), g())

    def testLength(self):
        u"""Differently sized iterators"""
        assert not Iter.equal(iter((1, 2, 3)), iter((1, 2)))
        assert not Iter.equal(iter((1, 2)), iter((1, 2, 3)))


class FilterTestCase(Iterators):
    u"""Tests for lazy_filter function"""
    def testEmpty(self):
        u"""empty iterators -> empty iterators"""
        assert Iter.empty(Iter.filter(self.alwayserror,
                                      self.empty())), \
            u"Filtering an empty iterator should result in empty iterator"

    def testNum1(self):
        u"""Test numbers 1 - 100 #1"""
        assert Iter.equal(Iter.filter(lambda x: x % 2 == 0,
                                      self.one_to_100()),
                          self.evens())
        assert Iter.equal(Iter.filter(lambda x: x % 2,
                                      self.one_to_100()),
                          self.odds())

    def testError(self):
        u"""Should raise appropriate error"""
        i = Iter.filter(lambda x: x, self.falseerror_maker())
        if sys.version_info.major >= 3:
            self.assertRaises(Exception, i.__next__)
        else:
            self.assertRaises(Exception, i.next)


class MapTestCase(Iterators):
    u"""Test mapping of iterators"""
    def testNumbers(self):
        u"""1 to 100 * 2 = 2 to 200"""
        assert Iter.equal(Iter.map(lambda x: 2 * x, self.one_to_100()),
                          iter(list(range(2, 201, 2))))

    def testShortcut(self):
        u"""Map should go in order"""
        def f(x):
            if x == u"hello":
                raise NameError
        i = Iter.map(f, self.trueerror_maker())
        next(i)
        if sys.version_info.major >= 3:
            self.assertRaises(NameError, i.__next__)
        else:
            self.assertRaises(NameError, i.next)

    def testEmpty(self):
        u"""Map of an empty iterator is empty"""
        assert Iter.empty(Iter.map(lambda x: x, iter([])))


class CatTestCase(Iterators):
    u"""Test concatenation of iterators"""
    def testEmpty(self):
        u"""Empty + empty = empty"""
        assert Iter.empty(Iter.cat(iter([]), iter([])))

    def testNumbers(self):
        u"""1 to 50 + 51 to 100 = 1 to 100"""
        assert Iter.equal(Iter.cat(iter(list(range(1, 51))), iter(list(range(51, 101)))),
                          self.one_to_100())

    def testShortcut(self):
        u"""Process iterators in order"""
        i = Iter.cat(self.typeerror_maker(), self.nameerror_maker())
        next(i)
        next(i)
        if sys.version_info.major >= 3:
            self.assertRaises(TypeError, i.__next__)
        else:
            self.assertRaises(TypeError, i.next)


class AndOrTestCase(Iterators):
    u"""Test And and Or"""
    def testEmpty(self):
        u"""And() -> true, Or() -> false"""
        assert Iter.And(self.empty())
        assert not Iter.Or(self.empty())

    def testAndShortcut(self):
        u"""And should return if any false"""
        assert Iter.And(self.falseerror_maker()) is None

    def testOrShortcut(self):
        u"""Or should return if any true"""
        assert Iter.Or(self.trueerror_maker()) == 1

    def testNormalAnd(self):
        u"""And should go through true iterators, picking last"""
        assert Iter.And(iter([1, 2, 3, 4])) == 4
        self.assertRaises(Exception, Iter.And, self.trueerror_maker())

    def testNormalOr(self):
        u"""Or goes through false iterators, picking last"""
        assert Iter.Or(iter([0, None, []])) == []
        self.assertRaises(Exception, Iter.Or, self.falseerror_maker())


class FoldingTest(Iterators):
    u"""Test folding operations"""
    def f(self, x, y):
        return x + y

    def testEmpty(self):
        u"""Folds of empty iterators should produce defaults"""
        assert Iter.foldl(self.f, 23, self.empty()) == 23
        assert Iter.foldr(self.f, 32, self.empty()) == 32

    def testAddition(self):
        u"""Use folds to sum lists"""
        assert Iter.foldl(self.f, 0, self.one_to_100()) == 5050
        assert Iter.foldr(self.f, 0, self.one_to_100()) == 5050

    def testLargeAddition(self):
        u"""Folds on 10000 element iterators"""
        assert Iter.foldl(self.f, 0, iter(list(range(1, 10001)))) == 50005000
        self.assertRaises(RuntimeError,
                          Iter.foldr, self.f, 0, iter(list(range(1, 10001))))

    def testLen(self):
        u"""Use folds to calculate length of lists"""
        assert Iter.foldl(lambda x, y: x + 1, 0, self.evens()) == 50
        assert Iter.foldr(lambda x, y: y + 1, 0, self.odds()) == 50


class MultiplexTest(Iterators):
    def testSingle(self):
        u"""Test multiplex single stream"""
        i_orig = self.one_to_100()
        i2_orig = self.one_to_100()
        i = Iter.multiplex(i_orig, 1)[0]
        assert Iter.equal(i, i2_orig)

    def testTrible(self):
        u"""Test splitting iterator into three"""
        counter = [0]

        def ff(x):
            counter[0] += 1

        i_orig = self.one_to_100()
        i2_orig = self.one_to_100()
        i1, i2, i3 = Iter.multiplex(i_orig, 3, ff)
        assert Iter.equal(i1, i2)
        assert Iter.equal(i3, i2_orig)
        assert counter[0] == 100, counter

    def testDouble(self):
        u"""Test splitting into two..."""
        i1, i2 = Iter.multiplex(self.one_to_100(), 2)
        assert Iter.equal(i1, self.one_to_100())
        assert Iter.equal(i2, self.one_to_100())


class ITRBadder(ITRBranch):
    def start_process(self, index):
        self.total = 0

    def end_process(self):
        if self.base_index:
            summand = self.base_index[-1]
            # print "Adding ", summand
            self.total += summand

    def branch_process(self, subinstance):
        # print "Adding subinstance ", subinstance.total
        self.total += subinstance.total


class ITRBadder2(ITRBranch):
    def start_process(self, index):
        self.total = 0

    def end_process(self):
        # print "Adding ", self.base_index
        self.total += reduce(lambda x, y: x + y, self.base_index, 0)

    def can_fast_process(self, index):
        if len(index) == 3:
            return 1
        else:
            return None

    def fast_process(self, index):
        self.total += index[0] + index[1] + index[2]

    def branch_process(self, subinstance):
        # print "Adding branch ", subinstance.total
        self.total += subinstance.total


class TreeReducerTest(UnitTestCase):
    def setUp(self):
        super(TreeReducerTest, self).setUp()

        self.i1 = [(), (1,), (2,), (3,)]
        self.i2 = [(0,), (0, 1), (0, 1, 0), (0, 1, 1), (0, 2), (0, 2, 1), (0, 3)]

        self.i1a = [(), (1,)]
        self.i1b = [(2,), (3,)]
        self.i2a = [(0,), (0, 1), (0, 1, 0)]
        self.i2b = [(0, 1, 1), (0, 2)]
        self.i2c = [(0, 2, 1), (0, 3)]

    def testTreeReducer(self):
        u"""testing IterTreeReducer"""
        itm = IterTreeReducer(ITRBadder, [])
        for index in self.i1:
            val = itm(index)
            assert val, (val, index)
        itm.Finish()
        assert itm.root_branch.total == 6, itm.root_branch.total

        itm2 = IterTreeReducer(ITRBadder2, [])
        for index in self.i2:
            val = itm2(index)
            if index == ():
                assert not val
            else:
                assert val
        itm2.Finish()
        assert itm2.root_branch.total == 12, itm2.root_branch.total

    def testTreeReducerState(self):
        u"""Test saving and recreation of an IterTreeReducer"""
        itm1a = IterTreeReducer(ITRBadder, [])
        for index in self.i1a:
            val = itm1a(index)
            assert val, index
        itm1b = pickle.loads(pickle.dumps(itm1a))
        for index in self.i1b:
            val = itm1b(index)
            assert val, index
        itm1b.Finish()
        assert itm1b.root_branch.total == 6, itm1b.root_branch.total

        itm2a = IterTreeReducer(ITRBadder2, [])
        for index in self.i2a:
            val = itm2a(index)
            if index == ():
                assert not val
            else:
                assert val
        itm2b = pickle.loads(pickle.dumps(itm2a))
        for index in self.i2b:
            val = itm2b(index)
            if index == ():
                assert not val
            else:
                assert val
        itm2c = pickle.loads(pickle.dumps(itm2b))
        for index in self.i2c:
            val = itm2c(index)
            if index == ():
                assert not val
            else:
                assert val
        itm2c.Finish()
        assert itm2c.root_branch.total == 12, itm2c.root_branch.total


if __name__ == u"__main__":
    unittest.main()



from __future__ import print_function
from future import standard_library
standard_library.install_aliases()
from builtins import filter
from builtins import map
from builtins import next
from builtins import object
from builtins import range



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2014 Andrea Grandi <a.grandi@gmail.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

import os.path
import duplicity.backend
import duplicity.util


class SXBackend(duplicity.backend.Backend):
    u"""Connect to remote store using Skylable Protocol"""
    def __init__(self, parsed_url):
        duplicity.backend.Backend.__init__(self, parsed_url)
        self.url_string = parsed_url.url_string

    def _put(self, source_path, remote_filename):
        remote_filename = util.fsdecode(remote_filename)
        remote_path = os.path.join(self.url_string, remote_filename)
        commandline = u"sxcp {0} {1}".format(source_path.uc_name, remote_path)
        self.subprocess_popen(commandline)

    def _get(self, remote_filename, local_path):
        remote_filename = util.fsdecode(remote_filename)
        remote_path = os.path.join(self.url_string, remote_filename)
        commandline = u"sxcp {0} {1}".format(remote_path, local_path.uc_name)
        self.subprocess_popen(commandline)

    def _list(self):
        # Do a long listing to avoid connection reset
        commandline = u"sxls {0}".format(self.url_string)
        _, l, _ = self.subprocess_popen(commandline)
        # Look for our files as the last element of a long list line
        return [util.fsencode(x[x.rindex(u'/') + 1:].split()[-1]) for x in l.split(u'\n')
                if x and not x.startswith(u"total ")]

    def _delete(self, filename):
        commandline = u"sxrm {0}/{1}".format(self.url_string, filename)
        self.subprocess_popen(commandline)


duplicity.backend.register_backend(u"sx", SXBackend)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2014 Canonical Ltd
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

from __future__ import print_function
from future import standard_library
standard_library.install_aliases()

import os
import io
import sys
import unittest

import duplicity.backend
from duplicity import log
from duplicity import globals
from duplicity import path
from duplicity import util
from duplicity.errors import BackendException
from . import UnitTestCase


class BackendInstanceBase(UnitTestCase):

    def setUp(self):
        UnitTestCase.setUp(self)
        assert not os.system(u"rm -rf testfiles")
        os.makedirs(u'testfiles')
        self.backend = None
        self.local = path.Path(u'testfiles/local')
        self.local.writefileobj(io.BytesIO(b"hello"))

    def tearDown(self):
        assert not os.system(u"rm -rf testfiles")
        if self.backend is None:
            return
        if hasattr(self.backend, u'_close'):
            self.backend._close()

    def test_get(self):
        if self.backend is None:
            return
        self.backend._put(self.local, b'a')
        getfile = path.Path(u'testfiles/getfile')
        self.backend._get(b'a', getfile)
        self.assertTrue(self.local.compare_data(getfile))

    def test_list(self):
        if self.backend is None:
            return
        self.backend._put(self.local, b'a')
        self.backend._put(self.local, b'b')
        # It's OK for backends to create files as a side effect of put (e.g.
        # the par2 backend does), so only check that at least a and b exist.
        self.assertTrue(b'a' in self.backend._list())
        self.assertTrue(b'b' in self.backend._list())

    def test_delete(self):
        if self.backend is None:
            return
        if not hasattr(self.backend, u'_delete'):
            self.assertTrue(hasattr(self.backend, u'_delete_list'))
            return
        self.backend._put(self.local, b'a')
        self.backend._put(self.local, b'b')
        self.backend._delete(b'a')
        self.assertFalse(b'a' in self.backend._list())
        self.assertTrue(b'b' in self.backend._list())

    def test_delete_clean(self):
        if self.backend is None:
            return
        if not hasattr(self.backend, u'_delete'):
            self.assertTrue(hasattr(self.backend, u'_delete_list'))
            return
        self.backend._put(self.local, b'a')
        self.backend._delete(b'a')
        self.assertEqual(self.backend._list(), [])

    def test_delete_missing(self):
        if self.backend is None:
            return
        if not hasattr(self.backend, u'_delete'):
            self.assertTrue(hasattr(self.backend, u'_delete_list'))
            return
        # Backends can either silently ignore this, or throw an error
        # that gives log.ErrorCode.backend_not_found.
        try:
            self.backend._delete(b'a')
        except BackendException as e:
            pass  # Something went wrong, but it was an 'expected' something
        except Exception as e:
            code = duplicity.backend._get_code_from_exception(self.backend, u'delete', e)
            self.assertEqual(code, log.ErrorCode.backend_not_found)

    def test_delete_list(self):
        if self.backend is None:
            return
        if not hasattr(self.backend, u'_delete_list'):
            self.assertTrue(hasattr(self.backend, u'_delete'))
            return
        self.backend._put(self.local, b'a')
        self.backend._put(self.local, b'b')
        self.backend._put(self.local, b'c')
        self.backend._delete_list([b'a', b'd', b'c'])
        files = self.backend._list()
        self.assertFalse(b'a' in files, files)
        self.assertTrue(b'b' in files, files)
        self.assertFalse(b'c' in files, files)

    def test_move(self):
        if self.backend is None:
            return
        if not hasattr(self.backend, u'_move'):
            return

        copy = path.Path(u'testfiles/copy')
        self.local.copy(copy)

        self.backend._move(self.local, b'a')
        self.assertTrue(b'a' in self.backend._list())
        self.assertFalse(self.local.exists())

        getfile = path.Path(u'testfiles/getfile')
        self.backend._get(b'a', getfile)
        self.assertTrue(copy.compare_data(getfile))

    def test_query_exists(self):
        if self.backend is None:
            return
        if not hasattr(self.backend, u'_query'):
            return
        self.backend._put(self.local, b'a')
        info = self.backend._query(b'a')
        self.assertEqual(info[u'size'], self.local.getsize())

    def test_query_missing(self):
        if self.backend is None:
            return
        if not hasattr(self.backend, u'_query'):
            return
        # Backends can either return -1 themselves, or throw an error
        # that gives log.ErrorCode.backend_not_found.
        try:
            info = self.backend._query(b'a')
        except BackendException as e:
            pass  # Something went wrong, but it was an 'expected' something
        except Exception as e:
            code = duplicity.backend._get_code_from_exception(self.backend, u'query', e)
            self.assertEqual(code, log.ErrorCode.backend_not_found)
        else:
            self.assertEqual(info[u'size'], -1)

    def test_query_list(self):
        if self.backend is None:
            return
        if not hasattr(self.backend, u'_query_list'):
            return
        self.backend._put(self.local, b'a')
        self.backend._put(self.local, b'c')
        info = self.backend._query_list([b'a', b'b'])
        self.assertEqual(info[b'a'][u'size'], self.local.getsize())
        self.assertEqual(info[b'b'][u'size'], -1)
        self.assertFalse(b'c' in info)


class LocalBackendTest(BackendInstanceBase):
    def setUp(self):
        super(LocalBackendTest, self).setUp()
        url = u'file://testfiles/output'
        self.backend = duplicity.backend.get_backend_object(url)
        self.assertEqual(self.backend.__class__.__name__, u'LocalBackend')


# TODO: Add par2-specific tests here, to confirm that we can recover
class Par2BackendTest(BackendInstanceBase):
    def setUp(self):
        super(Par2BackendTest, self).setUp()
        url = u'par2+file://testfiles/output'
        self.backend = duplicity.backend.get_backend_object(url)
        self.assertEqual(self.backend.__class__.__name__, u'Par2Backend')


# TODO: Fix duplicity.errors.InvalidBackendURL: Missing hostname ...
#  class RsyncBackendTest(BackendInstanceBase):
#      def setUp(self):
#          super(RsyncBackendTest, self).setUp()
#          os.makedirs('testfiles/output')  # rsync needs it to exist first
#          url = 'rsync://%s/testfiles/output' % os.getcwd()
#          self.backend = duplicity.backend.get_backend_object(url)
#          self.assertEqual(self.backend.__class__.__name__, 'RsyncBackend')


class TahoeBackendTest(BackendInstanceBase):
    def setUp(self):
        super(TahoeBackendTest, self).setUp()
        os.makedirs(u'testfiles/output')
        url = u'tahoe://testfiles/output'
        self.backend = duplicity.backend.get_backend_object(url)
        self.assertEqual(self.backend.__class__.__name__, u'TAHOEBackend')


# TODO: Modernize hsi backend stub
#  class HSIBackendTest(BackendInstanceBase):
#      def setUp(self):
#          super(HSIBackendTest, self).setUp()
#          os.makedirs(u'testfiles/output')
#          # hostname is ignored...  Seemingly on purpose
#          url = u'hsi://hostname%s/testfiles/output' % os.getcwd()
#          self.backend = duplicity.backend.get_backend_object(url)
#          self.assertEqual(self.backend.__class__.__name__, u'HSIBackend')


class FTPBackendTest(BackendInstanceBase):
    def setUp(self):
        super(FTPBackendTest, self).setUp()
        os.makedirs(u'testfiles/output')
        url = u'ftp://user:pass@hostname/testfiles/output'
        self.backend = duplicity.backend.get_backend_object(url)
        self.assertEqual(self.backend.__class__.__name__, u'LFTPBackend')


class FTPSBackendTest(BackendInstanceBase):
    def setUp(self):
        super(FTPSBackendTest, self).setUp()
        os.makedirs(u'testfiles/output')
        url = u'ftps://user:pass@hostname/testfiles/output'
        self.backend = duplicity.backend.get_backend_object(url)
        self.assertEqual(self.backend.__class__.__name__, u'LFTPBackend')


@unittest.skipIf(not util.which(u'rclone'), u"rclone not installed")
class RCloneBackendTest(BackendInstanceBase):
    def setUp(self):
        super(RCloneBackendTest, self).setUp()
        os.makedirs(u'testfiles/output')
        url = u'rclone://duptest:/%s/testfiles/output' % os.getcwd()
        self.backend = duplicity.backend.get_backend_object(url)
        self.assertEqual(self.backend.__class__.__name__, u'RcloneBackend')



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
#
# This file is part of duplicity.
#
# duplicity is free software; you can redistribute it and/or modify
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

u"""Provides a high-level interface to some librsync functions

This is a python wrapper around the lower-level _librsync module,
which is written in C.  The goal was to use C as little as possible...

"""

from builtins import str
from builtins import object
import os
import sys
from . import _librsync
import types
import array

if os.environ.get(u'READTHEDOCS') == u'True':
    import mock  # pylint: disable=import-error
    import duplicity
    duplicity._librsync = mock.MagicMock()

blocksize = _librsync.RS_JOB_BLOCKSIZE


class librsyncError(Exception):
    u"""Signifies error in internal librsync processing (bad signature, etc.)

    underlying _librsync.librsyncError's are regenerated using this
    class because the C-created exceptions are by default
    unPickleable.  There is probably a way to fix this in _librsync,
    but this scheme was easier.

    """
    pass


class LikeFile(object):
    u"""File-like object used by SigFile, DeltaFile, and PatchFile"""
    mode = u"rb"

    # This will be replaced in subclasses by an object with
    # appropriate cycle() method
    maker = None

    def __init__(self, infile, need_seek=None):
        u"""LikeFile initializer - zero buffers, set eofs off"""
        self.check_file(infile, need_seek)
        self.infile = infile
        self.closed = self.infile_closed = None
        self.inbuf = b""
        self.outbuf = array.array(u'b')
        self.eof = self.infile_eof = None

    def check_file(self, file, need_seek=None):
        u"""Raise type error if file doesn't have necessary attributes"""
        if not hasattr(file, u"read"):
            raise TypeError(u"Basis file must have a read() method")
        if not hasattr(file, u"close"):
            raise TypeError(u"Basis file must have a close() method")
        if need_seek and not hasattr(file, u"seek"):
            raise TypeError(u"Basis file must have a seek() method")

    def read(self, length=-1):
        u"""Build up self.outbuf, return first length bytes"""
        if length == -1:
            while not self.eof:
                self._add_to_outbuf_once()
            real_len = len(self.outbuf)
        else:
            while not self.eof and len(self.outbuf) < length:
                self._add_to_outbuf_once()
            real_len = min(length, len(self.outbuf))

        if sys.version_info.major >= 3:
            return_val = self.outbuf[:real_len].tobytes()
        else:
            return_val = self.outbuf[:real_len].tostring()
        del self.outbuf[:real_len]
        return return_val

    def _add_to_outbuf_once(self):
        u"""Add one cycle's worth of output to self.outbuf"""
        if not self.infile_eof:
            self._add_to_inbuf()
        try:
            self.eof, len_inbuf_read, cycle_out = self.maker.cycle(self.inbuf)
        except _librsync.librsyncError as e:
            raise librsyncError(str(e))
        self.inbuf = self.inbuf[len_inbuf_read:]
        if sys.version_info.major >= 3:
            self.outbuf.frombytes(cycle_out)
        else:
            self.outbuf.fromstring(cycle_out)

    def _add_to_inbuf(self):
        u"""Make sure len(self.inbuf) >= blocksize"""
        assert not self.infile_eof
        while len(self.inbuf) < blocksize:
            new_in = self.infile.read(blocksize)
            if not new_in:
                self.infile_eof = 1
                assert not self.infile.close()
                self.infile_closed = 1
                break
            self.inbuf += new_in

    def close(self):
        u"""Close infile"""
        if not self.infile_closed:
            assert not self.infile.close()
        self.closed = 1


class SigFile(LikeFile):
    u"""File-like object which incrementally generates a librsync signature"""
    def __init__(self, infile, blocksize=_librsync.RS_DEFAULT_BLOCK_LEN):
        u"""SigFile initializer - takes basis file

        basis file only needs to have read() and close() methods.  It
        will be closed when we come to the end of the signature.

        """
        LikeFile.__init__(self, infile)
        try:
            self.maker = _librsync.new_sigmaker(blocksize)
        except _librsync.librsyncError as e:
            raise librsyncError(str(e))


class DeltaFile(LikeFile):
    u"""File-like object which incrementally generates a librsync delta"""
    def __init__(self, signature, new_file):
        u"""DeltaFile initializer - call with signature and new file

        Signature can either be a string or a file with read() and
        close() methods.  New_file also only needs to have read() and
        close() methods.  It will be closed when self is closed.

        """
        LikeFile.__init__(self, new_file)
        if isinstance(signature, bytes):
            sig_string = signature
        else:
            self.check_file(signature)
            sig_string = signature.read()
            assert not signature.close()
        try:
            self.maker = _librsync.new_deltamaker(sig_string)
        except _librsync.librsyncError as e:
            raise librsyncError(str(e))


class PatchedFile(LikeFile):
    u"""File-like object which applies a librsync delta incrementally"""
    def __init__(self, basis_file, delta_file):
        u"""PatchedFile initializer - call with basis delta

        Here basis_file must be a true Python file, because we may
        need to seek() around in it a lot, and this is done in C.
        delta_file only needs read() and close() methods.

        """
        LikeFile.__init__(self, delta_file)
        try:
            basis_file.fileno()
        except:
            u""" tempfile.TemporaryFile() only guarantees a true file
            object on posix platforms. on cygwin/windows a file-like
            object whose file attribute is the underlying true file
            object is returned.
            """
            if hasattr(basis_file, u'file') and hasattr(basis_file.file, u'fileno'):
                basis_file = basis_file.file
            else:
                raise TypeError(_(u"basis_file must be a (true) file or an object whose "
                                  u"file attribute is the underlying true file object"))
        try:
            self.maker = _librsync.new_patchmaker(basis_file)
        except _librsync.librsyncError as e:
            raise librsyncError(str(e))


class SigGenerator(object):
    u"""Calculate signature.

    Input and output is same as SigFile, but the interface is like md5
    module, not filelike object

    """
    def __init__(self, blocksize=_librsync.RS_DEFAULT_BLOCK_LEN):
        u"""Return new signature instance"""
        try:
            self.sig_maker = _librsync.new_sigmaker(blocksize)
        except _librsync.librsyncError as e:
            raise librsyncError(str(e))
        self.gotsig = None
        self.buffer = b""
        self.sigstring_list = []

    def update(self, buf):
        u"""Add buf to data that signature will be calculated over"""
        if self.gotsig:
            raise librsyncError(u"SigGenerator already provided signature")
        self.buffer += buf
        while len(self.buffer) >= blocksize:
            if self.process_buffer():
                raise librsyncError(u"Premature EOF received from sig_maker")

    def process_buffer(self):
        u"""Run self.buffer through sig_maker, add to self.sig_string"""
        try:
            eof, len_buf_read, cycle_out = self.sig_maker.cycle(self.buffer)
        except _librsync.librsyncError as e:
            raise librsyncError(str(e))
        self.buffer = self.buffer[len_buf_read:]
        self.sigstring_list.append(cycle_out)
        return eof

    def getsig(self):
        u"""Return signature over given data"""
        while not self.process_buffer():
            pass  # keep running until eof
        return b''.join(self.sigstring_list)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf8 -*-
#
# Copyright 2014 Michael Terry <mike@mterry.name>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

# This is just a small override to the system gettext.py which allows us to
# always return a string with fancy unicode characters, which will notify us
# if we ever get a unicode->ascii translation by accident.


def install(*args, **kwargs):
    ZWSP = u"​"  # ZERO WIDTH SPACE, basically an invisible space separator
    import sys
    if sys.version_info.major >= 3:
        import builtins
        b = builtins
    else:
        import __builtin__
        b = __builtin__
    b.__dict__[u'_'] = lambda x: x + ZWSP
    b.__dict__[u'ngettext'] = lambda one, more, n: one + ZWSP if n == 1 else more + ZWSP



u"""py-unit tests for GnuPG

COPYRIGHT:

Copyright (C) 2001  Frank J. Tobin, ftobin@neverending.org

LICENSE:

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
or see http://www.gnu.org/copyleft/lesser.html
"""

from __future__ import print_function
from future import standard_library
standard_library.install_aliases()

import unittest

import tempfile
import sys

from duplicity import gpginterface

__author__ = u"Frank J. Tobin, ftobin@neverending.org"
__version__ = u"0.2.2"
__revision__ = u"$Id: GnuPGInterfacetest.py,v 1.11 2009/06/06 17:35:19 loafman Exp $"


class BasicTest(unittest.TestCase):
    u"""an initializer superclass"""

    def __init__(self, methodName=None):
        self.gnupg = gpginterface.GnuPG()
        unittest.TestCase.__init__(self, methodName)


class GnuPGTests(BasicTest):
    u"""Tests for GnuPG class"""

    def __init__(self, methodName=None):
        BasicTest.__init__(self, methodName)

        self.gnupg.passphrase = u"Three blind mice"
        self.gnupg.options.armor = 1
        self.gnupg.options.meta_interactive = 0
        self.gnupg.options.extra_args.append(u'--no-secmem-warning')

    def do_create_fh_operation(self, args, input,
                               passphrase=None):
        creations = [u'stdin', u'stdout']

        # Make sure we're getting the passphrase to GnuPG
        # somehow!
        assert passphrase is not None or self.gnupg.passphrase is not None, \
            u"No way to send the passphrase to GnuPG!"

        # We'll handle the passphrase manually
        if passphrase is not None:
            creations.append(u'passphrase')

        proc = self.gnupg.run(args, create_fhs=creations)

        if passphrase is not None:
            proc.handles[u'passphrase'].write(passphrase)
            proc.handles[u'passphrase'].close()

        proc.handles[u'stdin'].write(input)
        proc.handles[u'stdin'].close()

        ciphertext = proc.handles[u'stdout'].read()
        proc.handles[u'stdout'].close()

        # Checking to make sure GnuPG exited successfully
        proc.wait()

        return ciphertext

    def do_attach_fh_operation(self, args, stdin, stdout,
                               passphrase=None):

        # Make sure we're getting the passphrase to GnuPG
        # somehow!
        assert passphrase is not None or self.gnupg.passphrase is not None, \
            u"No way to send the passphrase to GnuPG!"

        creations = []
        # We'll handle the passphrase manually
        if passphrase is not None:
            proc.handles.append(u'passphrase')  # @UndefinedVariable

        attachments = {u'stdin': stdin, u'stdout': stdout}

        proc = self.gnupg.run(args, create_fhs=creations,
                              attach_fhs=attachments)

        if passphrase is not None:
            proc.handles[u'passphrase'].write(passphrase)
            proc.handles[u'passphrase'].close()

        # Checking to make sure GnuPG exited successfully
        proc.wait()

    def test_create_fhs_solely(self):
        u"""Do GnuPG operations using solely the create_fhs feature"""
        plaintext = b"Three blind mice"

        ciphertext = self.do_create_fh_operation([u'--symmetric'],
                                                 plaintext)

        decryption = self.do_create_fh_operation([u'--decrypt'],
                                                 ciphertext,
                                                 self.gnupg.passphrase)
        assert decryption == plaintext, \
            u"GnuPG decrypted output does not match original input"

    def test_attach_fhs(self):
        u"""Do GnuPG operations using the attach_fhs feature"""
        plaintext_source = __file__

        plainfile = open(plaintext_source, u"rb")
        temp1 = tempfile.TemporaryFile()
        temp2 = tempfile.TemporaryFile()

        self.do_attach_fh_operation([u'--symmetric'],
                                    stdin=plainfile, stdout=temp1)

        temp1.seek(0)

        self.do_attach_fh_operation([u'--decrypt'],
                                    stdin=temp1, stdout=temp2)

        plainfile.seek(0)
        temp2.seek(0)

        assert fh_cmp(plainfile, temp2), \
            u"GnuPG decrypted output does not match original input"


class OptionsTests(BasicTest):
    u"""Tests for Options class"""

    def __init__(self, methodName=None):
        BasicTest.__init__(self, methodName)
        self.reset_options()

    def reset_options(self):
        self.gnupg.options = gpginterface.Options()

    def option_to_arg(self, option):
        return u'--' + option.replace(u'_', u'-')

    def test_boolean_args(self):
        u"""test Options boolean options that they generate
        proper arguments"""

        booleans = [u'armor', u'no_greeting', u'no_verbose',
                    u'batch', u'always_trust', u'rfc1991',
                    u'quiet', u'openpgp', u'force_v3_sigs',
                    u'no_options', u'textmode']

        for option in booleans:
            self.reset_options()
            setattr(self.gnupg.options, option, 1)
            arg = self.option_to_arg(option)

            should_be = [arg]
            result = self.gnupg.options.get_args()

            assert should_be == result, \
                u"failure to set option '%s'; should be %s, but result is %s" \
                % (option, should_be, result)

    def test_string_args(self):
        u"""test Options string-taking options that they generate
        proper arguments"""

        strings = [u'homedir', u'default_key', u'comment', u'compress_algo',
                   u'options']

        string_value = u'test-argument'

        for option in strings:
            self.reset_options()
            setattr(self.gnupg.options, option, string_value)
            arg = self.option_to_arg(option)

            should_be = [arg, string_value]
            result = self.gnupg.options.get_args()

            assert should_be == result, \
                u"failure to set option '%s'; should be %s, but result is %s" \
                % (option, should_be, result)

    def test_list_args(self):
        u"""test Options string-taking options that they generate
        proper arguments"""

        lists = [u'recipients', u'encrypt_to']
        list_value = [u'test1', u'test2']

        for option in lists:
            self.reset_options()
            setattr(self.gnupg.options, option, list_value)

            # special case for recipients, since their
            # respective argument is 'recipient', not 'recipients'
            if option == u'recipients':
                arg = u'--recipient'
            else:
                arg = self.option_to_arg(option)

            should_be = []
            for v in list_value:
                should_be.extend([arg, v])

            result = self.gnupg.options.get_args()

            assert should_be == result, \
                u"failure to set option '%s'; should be %s, but result is %s" \
                % (option, should_be, result)


class PipesTests(unittest.TestCase):
    u"""Tests for Pipes class"""

    def test_constructor(self):
        self.pipe = gpginterface.Pipe(1, 2, 0)
        assert self.pipe.parent == 1
        assert self.pipe.child == 2
        assert not self.pipe.direct

########################################################################


def fh_cmp(f1, f2, bufsize=8192):
    while 1:
        b1 = f1.read(bufsize)
        b2 = f2.read(bufsize)
        if b1 != b2:
            return 0
        if not b1:
            return 1

########################################################################

if __name__ == u"__main__":
    unittest.main()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# @author: Juan Antonio Moya Vicen <juan@nowcomputing.com>
#
u"""
Functions to compute progress of compress & upload files
The heuristics try to infer the ratio between the amount of data collected
by the deltas and the total size of the changing files. It also infers the
compression and encryption ration of the raw deltas before sending them to
the backend.
With the inferred ratios, the heuristics estimate the percentage of completion
and the time left to transfer all the (yet unknown) amount of data to send.
This is a forecast based on gathered evidence.
"""

from __future__ import absolute_import
from __future__ import division

from builtins import object
import collections as sys_collections
import math
import threading
import time
from datetime import datetime, timedelta
from duplicity import globals
from duplicity import log
import pickle
import os

tracker = None
progress_thread = None


class Snapshot(sys_collections.deque):
    u"""
    A convenience class for storing snapshots in a space/timing efficient manner
    Stores up to 10 consecutive progress snapshots, one for each volume
    """

    @staticmethod
    def unmarshall():
        u"""
        De-serializes cached data it if present
        """
        snapshot = Snapshot()
        # If restarting Full, discard marshalled data and start over
        if globals.restart is not None and globals.restart.start_vol >= 1:
            try:
                progressfd = open(u'%s/progress' % globals.archive_dir_path.name, u'r')
                snapshot = pickle.load(progressfd)
                progressfd.close()
            except:
                log.Warn(u"Warning, cannot read stored progress information from previous backup",
                         log.WarningCode.cannot_stat)
                snapshot = Snapshot()
        # Reached here no cached data found or wrong marshalling
        return snapshot

    def marshall(self):
        u"""
        Serializes object to cache
        """
        progressfd = open(b'%s/progress' % globals.archive_dir_path.name, u'wb+')
        pickle.dump(self, progressfd)
        progressfd.close()

    def __init__(self, iterable=[], maxlen=10):
        super(Snapshot, self).__init__(iterable, maxlen)
        self.last_vol = 0

    def get_snapshot(self, volume):
        nitems = len(self)
        if nitems <= 0:
            return 0.0
        return self[max(0, min(nitems + volume - self.last_vol - 1, nitems - 1))]

    def push_snapshot(self, volume, snapshot_data):
        self.append(snapshot_data)
        self.last_vol = volume

    def pop_snapshot(self):
        return self.popleft()

    def clear(self):
        super(Snapshot, self).clear()
        self.last_vol = 0


class ProgressTracker(object):

    def __init__(self):
        self.total_stats = None
        self.nsteps = 0
        self.start_time = None
        self.change_mean_ratio = 0.0
        self.change_r_estimation = 0.0
        self.progress_estimation = 0.0
        self.time_estimation = 0
        self.total_bytecount = 0
        self.last_total_bytecount = 0
        self.last_bytecount = 0
        self.stall_last_time = None
        self.last_time = None
        self.elapsed_sum = timedelta()
        self.speed = 0.0
        self.transfers = sys_collections.deque()
        self.is_full = False
        self.current_estimation = 0.0
        self.prev_estimation = 0.0
        self.prev_data = None

    def snapshot_progress(self, volume):
        u"""
        Snapshots the current progress status for each volume into the disk cache
        If backup is interrupted, next restart will deserialize the data and try start
        progress from the snapshot
        """
        if self.prev_data is not None:
            self.prev_data.push_snapshot(volume, self.progress_estimation)
            self.prev_data.marshall()

    def has_collected_evidence(self):
        u"""
        Returns true if the progress computation is on and duplicity has not
        yet started the first dry-run pass to collect some information
        """
        return (self.total_stats is not None)

    def log_upload_progress(self):
        u"""
        Aproximative and evolving method of computing the progress of upload
        """
        if not globals.progress or not self.has_collected_evidence():
            return

        current_time = datetime.now()
        if self.start_time is None:
            self.start_time = current_time
        if self.last_time is not None:
            elapsed = (current_time - self.last_time)
        else:
            elapsed = timedelta()
        self.last_time = current_time

        # Detect (and report) a stallment if no changing data for more than 5 seconds
        if self.stall_last_time is None:
            self.stall_last_time = current_time
        if (current_time - self.stall_last_time).seconds > max(5, 2 * globals.progress_rate):
            log.TransferProgress(100.0 * self.progress_estimation,
                                 self.time_estimation, self.total_bytecount,
                                 (current_time - self.start_time).seconds,
                                 self.speed,
                                 True
                                 )
            return

        self.nsteps += 1

        u"""
        Compute the ratio of information being written for deltas vs file sizes
        Using Knuth algorithm to estimate approximate upper bound in % of completion
        The progress is estimated on the current bytes written vs the total bytes to
        change as estimated by a first-dry-run. The weight is the ratio of changing
        data (Delta) against the total file sizes. (pessimistic estimation)
        The method computes the upper bound for the progress, when using a sufficient
        large volsize to accomodate and changes, as using a small volsize may inject
        statistical noise.
        """
        from duplicity import diffdir
        changes = diffdir.stats.NewFileSize + diffdir.stats.ChangedFileSize
        total_changes = self.total_stats.NewFileSize + self.total_stats.ChangedFileSize
        if total_changes == 0 or diffdir.stats.RawDeltaSize == 0:
            return

        # Snapshot current values for progress
        last_progress_estimation = self.progress_estimation

        if self.is_full:
            # Compute mean ratio of data transfer, assuming 1:1 data density
            self.current_estimation = float(self.total_bytecount) / float(total_changes)
        else:
            # Compute mean ratio of data transfer, estimating unknown progress
            change_ratio = float(self.total_bytecount) / float(diffdir.stats.RawDeltaSize)
            change_delta = change_ratio - self.change_mean_ratio
            self.change_mean_ratio += change_delta / float(self.nsteps)  # mean cumulated ratio
            self.change_r_estimation += change_delta * (change_ratio - self.change_mean_ratio)
            change_sigma = math.sqrt(math.fabs(self.change_r_estimation / float(self.nsteps)))

            u"""
            Combine variables for progress estimation
            Fit a smoothed curve that covers the most common data density distributions,
            aiming for a large number of incremental changes.
            The computation is:
                Use 50% confidence interval lower bound during first half of the progression.
                Conversely, use 50% C.I. upper bound during the second half. Scale it to the
                changes/total ratio
            """
            self.current_estimation = float(changes) / float(total_changes) * (
                (self.change_mean_ratio - 0.67 * change_sigma) * (1.0 - self.current_estimation) +
                (self.change_mean_ratio + 0.67 * change_sigma) * self.current_estimation
            )
            u"""
            In case that we overpassed the 100%, drop the confidence and trust more the mean as the
            sigma may be large.
            """
            if self.current_estimation > 1.0:
                self.current_estimation = float(changes) / float(total_changes) * (
                    (self.change_mean_ratio - 0.33 * change_sigma) * (1.0 - self.current_estimation) +
                    (self.change_mean_ratio + 0.33 * change_sigma) * self.current_estimation
                )
            u"""
            Meh!, if again overpassed the 100%, drop the confidence to 0 and trust only the mean.
            """
            if self.current_estimation > 1.0:
                self.current_estimation = self.change_mean_ratio * float(changes) / float(total_changes)

        u"""
        Lastly, just cap it... nothing else we can do to approximate it better.
        Cap it to 99%, as the remaining 1% to 100% we reserve for the last step
        uploading of signature and manifests
        """
        self.progress_estimation = max(0.0, min(self.prev_estimation +
                                                (1.0 - self.prev_estimation) *
                                                self.current_estimation, 0.99))

        u"""
        Estimate the time just as a projection of the remaining time, fit to a
        [(1 - x) / x] curve
        """
        # As sum of timedeltas, so as to avoid clock skew in long runs
        # (adding also microseconds)
        self.elapsed_sum += elapsed
        projection = 1.0
        if self.progress_estimation > 0:
            projection = (1.0 - self.progress_estimation) / self.progress_estimation
        self.time_estimation = int(projection * float(self.elapsed_sum.total_seconds()))

        # Apply values only when monotonic, so the estimates look more consistent to the human eye
        if self.progress_estimation < last_progress_estimation:
            self.progress_estimation = last_progress_estimation

        u"""
        Compute Exponential Moving Average of speed as bytes/sec of the last 30 probes
        """
        if elapsed.total_seconds() > 0:
            self.transfers.append(float(self.total_bytecount - self.last_total_bytecount) /
                                  float(elapsed.total_seconds()))
        self.last_total_bytecount = self.total_bytecount
        if len(self.transfers) > 30:
            self.transfers.popleft()
        self.speed = 0.0
        for x in self.transfers:
            self.speed = 0.3 * x + 0.7 * self.speed

        log.TransferProgress(100.0 * self.progress_estimation,
                             self.time_estimation,
                             self.total_bytecount,
                             (current_time - self.start_time).seconds,
                             self.speed,
                             False
                             )

    def annotate_written_bytes(self, bytecount):
        u"""
        Annotate the number of bytes that have been added/changed since last time
        this function was called.
        bytecount param will show the number of bytes since the start of the current
        volume and for the current volume
        """
        changing = max(bytecount - self.last_bytecount, 0)
        self.total_bytecount += int(changing)  # Annotate only changing bytes since last probe
        self.last_bytecount = bytecount
        if changing > 0:
            self.stall_last_time = datetime.now()

    def set_evidence(self, stats, is_full):
        u"""
        Stores the collected statistics from a first-pass dry-run, to use this
        information later so as to estimate progress
        """
        self.total_stats = stats
        self.is_full = is_full

    def set_start_volume(self, volume):
        self.prev_data = Snapshot.unmarshall()
        self.prev_estimation = self.prev_data.get_snapshot(volume)
        self.progress_estimation = max(0.0, min(self.prev_estimation, 0.99))

    def total_elapsed_seconds(self):
        u"""
        Elapsed seconds since the first call to log_upload_progress method
        """
        return (datetime.now() - self.start_time).seconds


def report_transfer(bytecount, totalbytes):
    u"""
    Method to call tracker.annotate_written_bytes from outside
    the class, and to offer the "function(long, long)" signature
    which is handy to pass as callback
    """
    global tracker
    global progress_thread
    if progress_thread is not None and tracker is not None:
        tracker.annotate_written_bytes(bytecount)


class LogProgressThread(threading.Thread):
    u"""
    Background thread that reports progress to the log,
    every --progress-rate seconds
    """
    def __init__(self):
        super(LogProgressThread, self).__init__()
        self.setDaemon(True)
        self.finished = False

    def run(self):
        global tracker
        if not globals.dry_run and globals.progress and tracker.has_collected_evidence():
            while not self.finished:
                tracker.log_upload_progress()
                time.sleep(globals.progress_rate)
